import os
import sys
sys.path.insert(0, os.path.abspath("."))

import json
from pathlib import Path

import pytest
import math

from idtap.classes.piece import (
    Piece,
    init_sec_categorization,
    durations_of_fixed_pitches,
)
from idtap.classes.phrase import Phrase
from idtap.classes.trajectory import Trajectory
from idtap.classes.pitch import Pitch
from idtap.classes.raga import Raga
from idtap.classes.articulation import Articulation
from idtap.classes.group import Group
from idtap.classes.chikari import Chikari
from idtap.classes.meter import Meter
from idtap.classes.assemblage import Assemblage
from idtap.enums import Instrument
from datetime import datetime


# Helper builders

def build_simple_piece():
    raga = Raga()
    t1 = Trajectory({'id': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    t2 = Trajectory({'id': 12, 'pitches': [Pitch()], 'dur_tot': 1})
    p1 = Phrase({'trajectories': [t1], 'dur_tot': 1, 'raga': raga})
    p2 = Phrase({'trajectories': [t2], 'dur_tot': 1, 'raga': raga})
    m1 = Meter([1], start_time=0, tempo=60)
    m2 = Meter([1], start_time=1, tempo=60)
    return Piece({'phrases': [p1, p2], 'raga': raga, 'meters': [m1, m2], 'instrumentation': [Instrument.Sitar]})


def build_simple_piece_full():
    raga = Raga({'fundamental': 240})
    art = {'0.00': Articulation({'stroke_nickname': 'da'})}
    t1 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 0.5, 'articulations': art})
    t2 = Trajectory({'num': 1, 'pitches': [Pitch({'swara': 'r', 'raised': False})], 'dur_tot': 0.5, 'articulations': art})
    group = Group({'trajectories': [t1, t2]})
    p1 = Phrase({'trajectories': [t1, t2], 'raga': raga})
    p1.groups_grid[0].append(group)
    t3 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    p2 = Phrase({'trajectories': [t3], 'raga': raga})
    piece = Piece({'phrases': [p1, p2], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    meter = Meter(start_time=0, tempo=60)
    return piece, p1, p2, t1, t2, t3, group, meter


# ---------------------------- tests ----------------------------

def test_realign_and_set_dur_tot():
    piece = build_simple_piece()
    piece.phrases[0].trajectories[0].pitches[0].ratios = [1]
    piece.realign_pitches()
    assert piece.phrases[0].trajectories[0].pitches[0].ratios[0] == piece.raga.stratified_ratios[0]

    piece.set_dur_tot(3)
    assert piece.dur_tot == 3
    assert pytest.approx(piece.dur_array_grid[0][0], rel=1e-6) == 1 / 3
    assert piece.phrases[1].trajectories[0].dur_tot == 2
    assert pytest.approx(piece.phrases[1].start_time, rel=1e-6) == 1


def test_dur_calculations_and_cleanup():
    piece = build_simple_piece()
    piece.phrases[0].trajectories[0].dur_tot = 2
    piece.phrases[0].dur_tot_from_trajectories()
    piece.dur_tot_from_phrases()
    assert piece.dur_tot == 3
    piece.dur_array_from_phrases()
    assert piece.dur_array_grid[0] == pytest.approx([2/3, 1/3])
    piece.phrases[0].dur_tot = None  # type: ignore
    with pytest.raises(Exception):
        piece.dur_array_from_phrases()

    c = init_sec_categorization()
    del c['Improvisation']
    del c['Other']
    del c['Top Level']
    c['Composition Type']['Bandish'] = True
    piece.clean_up_section_categorization(c)
    assert 'Improvisation' in c
    assert 'Other' in c
    assert c['Top Level'] == 'Composition'


def test_piece_serialization_round_trip(tmp_path: Path):
    fixture = Path('idtap/tests/fixtures/serialization_test.json')
    data = json.loads(fixture.read_text())
    piece = Piece.from_json(data)
    json_obj = piece.to_json()
    copy = Piece.from_json(json_obj)
    assert copy.to_json() == json_obj


def test_durations_and_proportions_each_type():
    raga = Raga()
    t1 = Trajectory({'id': 0, 'pitches': [Pitch({'swara': 0})], 'dur_tot': 1})
    t2 = Trajectory({'id': 0, 'pitches': [Pitch({'swara': 1})], 'dur_tot': 2})
    phrase = Phrase({'trajectories': [t1, t2], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})

    np1 = t1.pitches[0].numbered_pitch
    np2 = t2.pitches[0].numbered_pitch

    durPN = piece.durations_of_fixed_pitches()
    assert durPN == {np1: 1, np2: 2}

    propPN = piece.proportions_of_fixed_pitches()
    assert pytest.approx(propPN[np1]) == 1/3
    assert pytest.approx(propPN[np2]) == 2/3

    c1 = Pitch.pitch_number_to_chroma(np1)
    c2 = Pitch.pitch_number_to_chroma(np2)
    assert piece.durations_of_fixed_pitches(output_type='chroma') == {c1: 1, c2: 2}
    assert piece.proportions_of_fixed_pitches(output_type='chroma') == {c1: pytest.approx(1/3), c2: pytest.approx(2/3)}

    sd1 = Pitch.chroma_to_scale_degree(c1)[0]
    sd2 = Pitch.chroma_to_scale_degree(c2)[0]
    assert piece.durations_of_fixed_pitches(output_type='scaleDegree') == {sd1: 1, sd2: 2}
    assert piece.proportions_of_fixed_pitches(output_type='scaleDegree') == {sd1: pytest.approx(1/3), sd2: pytest.approx(2/3)}

    sarg1 = Pitch.from_pitch_number(np1).sargam_letter
    sarg2 = Pitch.from_pitch_number(np2).sargam_letter
    assert piece.durations_of_fixed_pitches(output_type='sargamLetter') == {sarg1: 1, sarg2: 2}
    assert piece.proportions_of_fixed_pitches(output_type='sargamLetter') == {sarg1: pytest.approx(1/3), sarg2: pytest.approx(2/3)}


def test_helper_durations_invalid_and_proportional():
    bad_traj = type('T', (), {'durations_of_fixed_pitches': lambda self, *_: 5})()
    with pytest.raises(SyntaxError):
        durations_of_fixed_pitches([bad_traj])

    t1 = Trajectory({'id':0,'pitches':[Pitch({'swara':0})],'dur_tot':1})
    t2 = Trajectory({'id':0,'pitches':[Pitch({'swara':1})],'dur_tot':2})
    np1 = t1.pitches[0].numbered_pitch
    np2 = t2.pitches[0].numbered_pitch
    result = durations_of_fixed_pitches([t1,t2], count_type='proportional')
    assert pytest.approx(result[np1]) == 1/3
    assert pytest.approx(result[np2]) == 2/3
    total = sum(result.values())
    assert pytest.approx(total) == 1


# -------------------------------------------------------
#  New tests mirroring additional Piece features
# -------------------------------------------------------

def build_multi_track_piece():
    raga = Raga({'fundamental': 200})
    tA1 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 0.5})
    tA2 = Trajectory({'num': 1, 'pitches': [Pitch()], 'dur_tot': 0.5})
    group = Group({'trajectories': [tA1, tA2]})
    pA = Phrase({'trajectories': [tA1, tA2], 'raga': raga})
    pA.groups_grid[0].append(group)

    tB1 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    pB = Phrase({'trajectories': [tB1], 'raga': raga})

    piece = Piece({
        'phraseGrid': [[pA], [pB]],
        'instrumentation': [Instrument.Sitar, Instrument.Vocal_M],
        'raga': raga,
    })

    return piece


def test_optional_fields_round_trip():
    opts = {
        'title': 'my title',
        'dateCreated': datetime(2020, 1, 1),
        'dateModified': datetime(2020, 1, 2),
        'location': 'home',
        '_id': 'id1',
        'audioID': 'a1',
        'audio_DB_ID': 'db1',
        'userID': 'u1',
        'name': 'name',
        'family_name': 'fam',
        'given_name': 'giv',
        'permissions': 'perm',
        'explicitPermissions': {'edit': ['e'], 'view': ['v'], 'publicView': False},
        'soloist': 'solo',
        'soloInstrument': 'sitar',
        'instrumentation': [Instrument.Sitar],
        'phrases': [],
        'raga': Raga(),
    }
    piece = Piece(opts)
    copy = Piece.from_json(piece.to_json())
    assert copy.title == opts['title']
    assert copy.date_created.isoformat() == opts['dateCreated'].isoformat()
    assert copy.date_modified.isoformat() == opts['dateModified'].isoformat()
    assert copy.location == opts['location']
    assert copy._id == opts['_id']
    assert copy.audio_id == opts['audioID']
    assert copy.user_id == opts['userID']
    assert copy.name == opts['name']
    assert copy.family_name == opts['family_name']
    assert copy.given_name == opts['given_name']
    assert copy.permissions == opts['permissions']
    assert copy.explicit_permissions == opts['explicitPermissions']
    assert copy.soloist == opts['soloist']
    assert copy.solo_instrument == opts['soloInstrument']


def test_getters_and_setters_modify_grids():
    raga = Raga()
    t1 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    p1 = Phrase({'trajectories': [t1], 'raga': raga})
    piece = Piece({'phrases': [p1], 'raga': raga, 'instrumentation': [Instrument.Sitar]})

    t2 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    p2 = Phrase({'trajectories': [t2], 'raga': raga})
    piece.phrases = [p2]
    assert piece.phrase_grid[0][0] is p2

    piece.durArray = [1]
    assert piece.dur_array_grid[0] == [1]

    piece.section_starts = [0]
    assert piece.section_starts_grid[0] == [0]

    sc = [init_sec_categorization()]
    piece.section_categorization = sc
    assert piece.section_cat_grid[0] is sc


def test_assemblages_getter():
    raga = Raga()
    traj = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    phrase = Phrase({'trajectories': [traj], 'raga': raga})
    asm = Assemblage(Instrument.Sitar, 'a')
    asm.add_phrase(phrase)
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    piece.assemblage_descriptors = [asm.descriptor]
    aggs = piece.assemblages
    assert len(aggs) == 1
    assert isinstance(aggs[0], Assemblage)
    assert aggs[0].phrases[0] is phrase


def test_update_start_times_recalc():
    raga = Raga()
    p1 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    p2 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    piece = Piece({'phrases': [p1, p2], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    piece.dur_array_grid[0] = [0.25, 0.75]
    piece.dur_tot = 2
    piece.update_start_times()
    assert pytest.approx(p2.start_time, rel=1e-6) == piece.dur_starts()[1]
    assert p2.piece_idx == 1


def test_track_specific_helpers():
    piece = build_multi_track_piece()
    assert piece.dur_starts(1) == [0]
    assert piece.traj_start_times(1) == [0]
    assert len(piece.all_pitches(track=1)) == 1
    traj = piece.phrase_grid[1][0].trajectories[0]
    assert piece.most_recent_traj(1.2, 1) is traj


def test_ad_hoc_grid_expansion():
    raga = Raga()
    traj = Trajectory({'dur_tot': 1})
    phrase = Phrase({'trajectories': [traj], 'raga': raga})
    piece = Piece({
        'phraseGrid': [[phrase]],
        'instrumentation': [Instrument.Sitar, Instrument.Vocal_M],
        'raga': raga,
        'adHocSectionCatGrid': [[]],
    })
    assert len(piece.ad_hoc_section_cat_grid) == 2


def test_section_cat_grid_expansion():
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    piece = Piece({
        'phrases': [phrase],
        'raga': raga,
        'instrumentation': [Instrument.Sitar],
        'sectionStarts': [0, 1],
        'sectionCatGrid': [[init_sec_categorization()]],
    })
    assert len(piece.section_cat_grid[0]) == 2


def test_add_trajectory_validation_basic():
    """Test basic input validation for add_trajectory."""
    piece = build_simple_piece()
    
    # Invalid track index
    assert not piece.add_trajectory({'id': 0, 'dur_tot': 0.5}, -1, 0.5)
    assert not piece.add_trajectory({'id': 0, 'dur_tot': 0.5}, 999, 0.5)
    
    # Invalid start time
    assert not piece.add_trajectory({'id': 0, 'dur_tot': 0.5}, 0, -1)
    
    # Invalid trajectory exceeding piece duration
    assert not piece.add_trajectory({'id': 0, 'dur_tot': 10}, 0, 0.5)


def test_add_trajectory_replace_entire_silent():
    """Test replacing an entire silent trajectory."""
    raga = Raga({'fundamental': 240})
    silent_traj = Trajectory({'id': 12, 'dur_tot': 2.0, 'fundID12': 240})
    phrase = Phrase({'trajectories': [silent_traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    
    # Create new trajectory to add 
    new_traj_data = {
        'id': 0,
        'dur_tot': 2.0,
        'pitches': [Pitch({'swara': 's', 'fundamental': 240})]
    }
    
    # Add trajectory that replaces entire silent trajectory
    success = piece.add_trajectory(new_traj_data, 0, 0.0)
    assert success
    
    # Verify replacement
    phrase = piece.phrase_grid[0][0]
    assert len(phrase.trajectories) == 1
    assert phrase.trajectories[0].id == 0
    assert phrase.trajectories[0].dur_tot == 2.0


def test_add_trajectory_replace_left_side():
    """Test replacing left side of silent trajectory."""
    raga = Raga({'fundamental': 240})
    silent_traj = Trajectory({'id': 12, 'dur_tot': 3.0, 'fundID12': 240})
    phrase = Phrase({'trajectories': [silent_traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    
    new_traj_data = {
        'id': 0,
        'dur_tot': 1.0,
        'pitches': [Pitch({'swara': 's', 'fundamental': 240})]
    }
    
    # Add trajectory at beginning of silent trajectory
    success = piece.add_trajectory(new_traj_data, 0, 0.0)
    assert success
    
    # Verify replacement
    phrase = piece.phrase_grid[0][0]
    assert len(phrase.trajectories) == 2
    assert phrase.trajectories[0].id == 0
    assert phrase.trajectories[0].dur_tot == 1.0
    assert phrase.trajectories[1].id == 12
    assert phrase.trajectories[1].dur_tot == 2.0


def test_add_trajectory_replace_right_side():
    """Test replacing right side of silent trajectory."""
    raga = Raga({'fundamental': 240})
    silent_traj = Trajectory({'id': 12, 'dur_tot': 3.0, 'fundID12': 240})
    phrase = Phrase({'trajectories': [silent_traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    
    new_traj_data = {
        'id': 0,
        'dur_tot': 1.0,
        'pitches': [Pitch({'swara': 's', 'fundamental': 240})]
    }
    
    # Add trajectory at end of silent trajectory
    success = piece.add_trajectory(new_traj_data, 0, 2.0)
    assert success
    
    # Verify replacement
    phrase = piece.phrase_grid[0][0]
    assert len(phrase.trajectories) == 2
    assert phrase.trajectories[0].id == 12
    assert phrase.trajectories[0].dur_tot == 2.0
    assert phrase.trajectories[1].id == 0
    assert phrase.trajectories[1].dur_tot == 1.0


def test_add_trajectory_replace_internal():
    """Test replacing internal portion of silent trajectory."""
    raga = Raga({'fundamental': 240})
    silent_traj = Trajectory({'id': 12, 'dur_tot': 5.0, 'fundID12': 240})
    phrase = Phrase({'trajectories': [silent_traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    
    new_traj_data = {
        'id': 0,
        'dur_tot': 2.0,
        'pitches': [Pitch({'swara': 's', 'fundamental': 240})]
    }
    
    # Add trajectory in middle of silent trajectory
    success = piece.add_trajectory(new_traj_data, 0, 1.5)
    assert success
    
    # Verify replacement - should create 3 trajectories: first silent, new, last silent
    phrase = piece.phrase_grid[0][0]
    assert len(phrase.trajectories) == 3
    assert phrase.trajectories[0].id == 12
    assert phrase.trajectories[0].dur_tot == 1.5
    assert phrase.trajectories[1].id == 0
    assert phrase.trajectories[1].dur_tot == 2.0
    assert phrase.trajectories[2].id == 12
    assert phrase.trajectories[2].dur_tot == 1.5


def test_add_trajectory_with_existing_trajectory_object():
    """Test adding pre-instantiated Trajectory object with different raga."""
    piece_raga = Raga({'fundamental': 240})
    different_raga = Raga({'fundamental': 220})
    
    silent_traj = Trajectory({'id': 12, 'dur_tot': 2.0, 'fundID12': 240})
    phrase = Phrase({'trajectories': [silent_traj], 'raga': piece_raga})
    piece = Piece({'phrases': [phrase], 'raga': piece_raga, 'instrumentation': [Instrument.Sitar]})
    
    # Create trajectory with different raga
    wrong_pitch = Pitch({'swara': 's', 'fundamental': 220, 'raga': different_raga})
    existing_traj = Trajectory({
        'id': 0,
        'dur_tot': 1.0,
        'pitches': [wrong_pitch],
        'instrumentation': Instrument.Vocal_M  # Also wrong instrumentation
    })
    
    # Add the pre-existing trajectory
    success = piece.add_trajectory(existing_traj, 0, 0.0)
    assert success
    
    # Verify the trajectory was recreated with piece's context
    added_traj = piece.phrase_grid[0][0].trajectories[0]
    assert added_traj.instrumentation == Instrument.Sitar  # Should match piece instrumentation
    # Verify pitches were updated with piece's raga fundamental
    for pitch in added_traj.pitches:
        assert pitch.fundamental == 240  # Should match piece's raga


def test_add_trajectory_validation_requirements():
    """Test the 5 validation requirements."""
    raga = Raga({'fundamental': 240})
    
    # Create piece with non-silent trajectory
    non_silent_traj = Trajectory({'id': 0, 'dur_tot': 2.0})
    phrase = Phrase({'trajectories': [non_silent_traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    
    new_traj_data = {'id': 1, 'dur_tot': 1.0}
    
    # Should fail - target trajectory is not silent (id != 12)
    assert not piece.add_trajectory(new_traj_data, 0, 0.0)
    
    # Create piece with silent trajectory for remaining tests
    silent_traj = Trajectory({'id': 12, 'dur_tot': 2.0, 'fundID12': 240})
    phrase2 = Phrase({'trajectories': [silent_traj], 'raga': raga})
    piece2 = Piece({'phrases': [phrase2], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    
    # Should fail - trajectory extends beyond piece duration
    long_traj = {'id': 0, 'dur_tot': 5.0}
    assert not piece2.add_trajectory(long_traj, 0, 0.0)
    
    # Should pass - valid trajectory within silent trajectory
    valid_traj = {'id': 0, 'dur_tot': 1.0}
    assert piece2.add_trajectory(valid_traj, 0, 0.5)


def test_add_trajectory_multi_track():
    """Test adding trajectories to different instrument tracks."""
    raga = Raga({'fundamental': 240})
    
    # Create piece with two tracks
    silent1 = Trajectory({'id': 12, 'dur_tot': 2.0, 'fundID12': 240})
    silent2 = Trajectory({'id': 12, 'dur_tot': 2.0, 'fundID12': 240})
    phrase1 = Phrase({'trajectories': [silent1], 'raga': raga})
    phrase2 = Phrase({'trajectories': [silent2], 'raga': raga})
    
    piece = Piece({
        'phraseGrid': [[phrase1], [phrase2]], 
        'raga': raga, 
        'instrumentation': [Instrument.Sitar, Instrument.Vocal_M]
    })
    
    # Add trajectory to first track
    traj1_data = {'id': 0, 'dur_tot': 1.0}
    success1 = piece.add_trajectory(traj1_data, 0, 0.0)
    assert success1
    
    # Add trajectory to second track
    traj2_data = {'id': 1, 'dur_tot': 1.0}  
    success2 = piece.add_trajectory(traj2_data, 1, 0.5)
    assert success2
    
    # Verify both tracks were modified correctly
    assert piece.phrase_grid[0][0].trajectories[0].id == 0
    assert piece.phrase_grid[0][0].trajectories[0].instrumentation == Instrument.Sitar
    
    assert piece.phrase_grid[1][0].trajectories[0].id == 12
    assert piece.phrase_grid[1][0].trajectories[1].id == 1
    assert piece.phrase_grid[1][0].trajectories[1].instrumentation == Instrument.Vocal_M


def test_all_pitches_no_repetition():
    raga = Raga()
    p1 = Pitch({'swara': 'sa'})
    t1 = Trajectory({'num': 0, 'phrase_idx': 0, 'pitches': [p1]})
    t2 = Trajectory({'num': 1, 'phrase_idx': 0, 'pitches': [Pitch({'swara': 'sa'})]})
    t3 = Trajectory({'num': 2, 'phrase_idx': 0, 'pitches': [Pitch({'swara': 're'})]})
    phrase = Phrase({'trajectories': [t1, t2, t3], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    assert len(piece.all_pitches()) == 3
    assert piece.all_pitches(repetition=False) == [p1, t3.pitches[0]]


def test_all_pitches_pitch_number_option():
    piece = build_simple_piece()
    nums = piece.all_pitches(pitch_number=True)
    assert all(isinstance(n, (int, float)) for n in nums)
    assert len(nums) > 0


def test_update_fundamental_and_chikari_freqs():
    piece, p1, *_ = build_simple_piece_full()
    base = piece.raga.fundamental
    assert piece.chikari_freqs(0) == [base * 2, base * 4]
    piece.update_fundamental(300)
    assert piece.raga.fundamental == 300
    assert p1.trajectories[0].pitches[0].fundamental == 300
    piece.put_raga_in_phrase()
    assert p1.raga is piece.raga
    chikari = Chikari({'fundamental': piece.raga.fundamental})
    p1.chikaris['0.00'] = chikari
    assert piece.chikari_freqs(0) == [c.frequency for c in chikari.pitches[:2]]
    nums = [p.numbered_pitch for p in piece.all_pitches()]
    assert piece.highest_pitch_number == max(nums)
    assert piece.lowest_pitch_number == min(nums)


def test_dur_starts_errors():
    piece = build_simple_piece()
    saved = piece.dur_array_grid
    piece.dur_array_grid = None
    with pytest.raises(Exception):
        piece.dur_starts()
    piece.dur_array_grid = saved
    piece.dur_tot = None
    with pytest.raises(Exception):
        piece.dur_starts()


def test_excerpt_range_and_assemblage_serialization():
    p = Phrase({'trajectories': [Trajectory()]})
    piece = Piece({'phrases': [p], 'raga': Raga()})
    piece.excerpt_range = {'start': 1, 'end': 2}
    asm = Assemblage(Instrument.Sitar, 'a')
    asm.add_phrase(p)
    piece.assemblage_descriptors = [asm.descriptor]
    copy = Piece.from_json(piece.to_json())
    assert copy.excerpt_range == piece.excerpt_range
    assert copy.assemblage_descriptors == piece.assemblage_descriptors


def test_dur_tot_and_permissions_persist():
    perms = {'edit': ['a'], 'view': ['b'], 'publicView': False}
    piece = Piece({'phrases': [], 'durTot': 5, 'instrumentation': [Instrument.Sitar], 'raga': Raga(), 'explicitPermissions': perms})
    # durTot is recalculated from phrases and becomes 0
    assert piece.dur_tot == 0
    assert piece.dur_array == []
    assert piece.explicit_permissions == perms
    assert piece.assemblage_descriptors == []


def test_clean_up_section_cat_defaults_multi_inst():
    sc = [init_sec_categorization()]
    del sc[0]['Improvisation']
    del sc[0]['Other']
    del sc[0]['Top Level']
    piece = Piece({'sectionStarts': [0], 'sectionCategorization': sc, 'instrumentation': [Instrument.Sitar, Instrument.Vocal_M], 'raga': Raga(), 'phrases': []})
    assert 'Improvisation' in sc[0]
    assert 'Other' in sc[0]
    assert 'Top Level' in sc[0]
    assert len(piece.ad_hoc_section_cat_grid) == 2
    assert piece.dur_tot == 0


# ----------------------------------------------------------------------
# Additional tests ported from the TypeScript suite
# ----------------------------------------------------------------------

def build_vocal_piece():
    raga = Raga({'fundamental': 240})
    art = {'0.00': Articulation({'stroke_nickname': 'da'})}
    t1 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 0.5, 'articulations': art})
    t1.add_consonant('ka')
    t1.update_vowel('a')
    t1.add_consonant('ga', start=False)
    t2 = Trajectory({'num': 1, 'pitches': [Pitch({'swara': 'r', 'raised': False})], 'dur_tot': 0.5, 'articulations': art})
    t2.update_vowel('i')
    p1 = Phrase({'trajectories': [t1, t2], 'raga': raga})
    p1.chikaris['0.25'] = Chikari({})
    p2 = Phrase({'trajectories': [Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})], 'raga': raga})
    piece = Piece({'phrases': [p1, p2], 'raga': raga, 'instrumentation': [Instrument.Vocal_M], 'sectionStarts': [0,1]})
    meter = Meter(start_time=0, tempo=60)
    piece.add_meter(meter)
    return piece, meter


def test_track_from_traj_uid_error():
    piece, *_ = build_simple_piece_full()
    with pytest.raises(ValueError):
        piece.track_from_traj_uid('missing')


def test_p_idx_from_group_across_phrases():
    raga = Raga()
    t1 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 0.5})
    t2 = Trajectory({'num': 1, 'pitches': [Pitch()], 'dur_tot': 0.5})
    g1 = Group({'trajectories': [t1, t2]})
    p1 = Phrase({'trajectories': [t1, t2], 'raga': raga})
    p1.groups_grid[0].append(g1)
    t3 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 0.5})
    t4 = Trajectory({'num': 1, 'pitches': [Pitch()], 'dur_tot': 0.5})
    g2 = Group({'trajectories': [t3, t4]})
    p2 = Phrase({'trajectories': [t3, t4], 'raga': raga})
    p2.groups_grid[0].append(g2)
    piece = Piece({'phrases': [p1, p2], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    assert piece.p_idx_from_group(g1) == 0
    assert piece.p_idx_from_group(g2) == 1


def test_most_recent_traj_and_chikari_freqs():
    piece, _ = build_vocal_piece()
    first_traj = piece.phrase_grid[0][0].trajectories[0]
    chikari = piece.phrase_grid[0][0].chikaris['0.25']
    assert piece.most_recent_traj(0.6, 0) is first_traj
    assert piece.chikari_freqs(0) == [c.frequency for c in chikari.pitches[:2]]


def test_add_meter_overlap_and_remove():
    raga = Raga()
    traj = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    phrase = Phrase({'trajectories': [traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    m1 = Meter(start_time=0, tempo=60)
    m2 = Meter(start_time=5, tempo=60)
    piece.add_meter(m1)
    piece.add_meter(m2)
    with pytest.raises(ValueError):
        piece.add_meter(Meter(start_time=3, tempo=60))
    piece.remove_meter(m1)
    assert piece.meters == [m2]


def test_comp_section_tempo_fallback():
    piece = build_simple_piece()
    c = init_sec_categorization()
    del c['Comp.-section/Tempo']
    c['Composition-section/Tempo'] = {'Madhya': True}
    del c['Top Level']
    piece.clean_up_section_categorization(c)
    assert c['Comp.-section/Tempo']['Madhya']
    assert 'Composition-section/Tempo' not in c
    assert c['Top Level'] == 'Composition'


@pytest.mark.parametrize('modify,expected', [
    (lambda c: c['Pre-Chiz Alap'].__setitem__('Pre-Chiz Alap', True), 'Pre-Chiz Alap'),
    (lambda c: c['Alap'].__setitem__('Alap', True), 'Alap'),
    (lambda c: c['Composition Type'].__setitem__('Bandish', True), 'Composition'),
    (lambda c: c['Comp.-section/Tempo'].__setitem__('Vilambit', True), 'Composition'),
    (lambda c: c['Improvisation'].__setitem__('Improvisation', True), 'Improvisation'),
    (lambda c: c['Other'].__setitem__('Other', True), 'Other'),
    (lambda c: None, 'None'),
])
def test_top_level_classification(modify, expected):
    piece = build_simple_piece()
    c = init_sec_categorization()
    del c['Top Level']
    modify(c)
    piece.clean_up_section_categorization(c)
    assert c['Top Level'] == expected


def test_all_display_vowels_non_vocal():
    raga = Raga()
    traj = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    phrase = Phrase({'trajectories': [traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    with pytest.raises(Exception, match='instrumentation is not vocal'):
        piece.all_display_vowels()


def test_all_pitches_number_error():
    raga = Raga()
    traj = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    traj.pitches.append(0)  # type: ignore
    phrase = Phrase({'trajectories': [traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    with pytest.raises(ValueError):
        piece.all_pitches(repetition=False)


def test_traj_from_time_after_last():
    piece = build_simple_piece()
    after = (piece.dur_tot or 0) + 1
    assert piece.traj_from_time(after, 0) is None


def test_traj_from_uid_error():
    piece = build_simple_piece()
    with pytest.raises(ValueError):
        piece.traj_from_uid('missing', 0)


def test_track_from_traj_error():
    piece, *_ = build_simple_piece_full()
    missing = Trajectory({'num': 99, 'pitches': [Pitch()], 'dur_tot': 1})
    with pytest.raises(ValueError):
        piece.track_from_traj(missing)


def test_phrase_from_uid_and_track_from_phrase_uid_error():
    piece, *_ = build_simple_piece_full()
    with pytest.raises(ValueError):
        piece.phrase_from_uid('missing')
    with pytest.raises(ValueError):
        piece.track_from_phrase_uid('missing')


def test_add_meter_enclosing_rejects():
    raga = Raga()
    traj = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    phrase = Phrase({'trajectories': [traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    base = Meter(start_time=0, tempo=60)
    piece.add_meter(base)
    with pytest.raises(ValueError):
        piece.add_meter(Meter(start_time=-1, hierarchy=[8], tempo=60))
    later = Meter(start_time=5, tempo=60)
    piece.add_meter(later)
    assert piece.meters == [base, later]


def test_dur_array_branch_empty_section_cat_grid():
    raga = Raga()
    traj = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 1})
    phrase = Phrase({'trajectories': [traj], 'raga': raga})
    piece = Piece({'phrases': [phrase], 'durTot': 1, 'durArray': [1], 'instrumentation': [Instrument.Sitar], 'raga': raga, 'sectionStarts': [0], 'sectionCatGrid': []})
    assert piece.dur_array_grid == [[1]]
    assert len(piece.section_cat_grid) == 1
    assert len(piece.section_cat_grid[0]) == len(piece.section_starts_grid[0])


def test_dur_tot_from_phrases_adds_silent_phrase():
    raga = Raga()
    t1 = Trajectory({'dur_tot': 1})
    p1 = Phrase({'trajectories': [t1], 'raga': raga})
    piece = Piece({'phraseGrid': [[p1], []], 'instrumentation': [Instrument.Sitar, Instrument.Sitar], 'raga': raga})
    piece.dur_tot_from_phrases()
    assert len(piece.phrase_grid[1]) == 1
    silent = piece.phrase_grid[1][0].trajectories[0]
    assert silent.id == 12
    assert pytest.approx(silent.dur_tot) == 1


def test_dur_array_from_phrases_removes_nan():
    raga = Raga()
    piece = Piece({'raga': raga, 'instrumentation': [Instrument.Sitar]})
    good = Trajectory({'dur_tot': 1})
    bad = Trajectory({'dur_tot': float('nan')})
    phrase = Phrase({'trajectories': [good, bad], 'raga': raga})
    phrase.dur_tot_from_trajectories()
    piece.phrase_grid[0].append(phrase)
    assert math.isnan(phrase.dur_tot)
    piece.dur_array_from_phrases()
    assert len(phrase.trajectories) == 1
    assert pytest.approx(phrase.dur_tot) == 1


def test_display_helpers_and_meters():
    piece = build_simple_piece()

    divs = piece.all_phrase_divs()
    assert len(divs) == 1
    assert pytest.approx(divs[0]['time'], rel=1e-6) == 1

    div_chunks = piece.chunked_phrase_divs(0, 1)
    assert len(div_chunks) == 2
    assert len(div_chunks[0]) == 0
    assert len(div_chunks[1]) == 1

    sargam = piece.all_display_sargam()
    assert sargam[0]['sargam'] is not None
    sargam_chunks = piece.chunked_display_sargam(0, 1)
    assert len(sargam_chunks) == 2
    assert sum(len(c) for c in sargam_chunks) == len(sargam)

    meter_chunks = piece.chunked_meters(1)
    assert len(meter_chunks) == 2
    assert meter_chunks[0][0].start_time == 0
    assert meter_chunks[1][0].start_time == 1

    with pytest.raises(ValueError):
        piece.add_meter(Meter([1], tempo=60, start_time=0.5))


def test_piece_method_helpers():
    piece, p1, p2, t1, t2, t3, group, meter = build_simple_piece_full()

    chunks = piece.chunked_trajs(0, 1)
    assert len(chunks[0]) == 2
    assert len(chunks[1]) == 1

    chunks_small = piece.chunked_trajs(0, 0.75)
    assert len(chunks_small) == 3
    assert len(chunks_small[0]) == 2
    assert len(chunks_small[1]) == 2
    assert len(chunks_small[2]) == 1

    bols = piece.all_display_bols()
    assert len(bols) > 0
    assert len(piece.chunked_display_bols(0, 1)[0]) == len([b for b in bols if b['time'] < 1])

    piece.add_meter(meter)
    pid = meter.all_pulses[0].unique_id
    assert piece.pulse_from_id(pid) == meter.all_pulses[0]


def test_vocal_display_helpers():
    piece, meter = build_vocal_piece()

    assert len(piece.sections) == 2
    assert len(piece.sections_grid[0]) == 2

    vowels = piece.all_display_vowels()
    assert len(vowels) > 0
    assert len(piece.chunked_display_vowels(0, 1)[0]) == len([v for v in vowels if v['time'] < 1])

    cons = piece.all_display_ending_consonants()
    assert len(cons) > 0
    assert len(piece.chunked_display_consonants(0, 1)[0]) == len([c for c in cons if c['time'] < 1])

    chiks = piece.all_display_chikaris()
    assert len(chiks) > 0
    assert len(piece.chunked_display_chikaris(0, 1)[0]) == len([c for c in chiks if c['time'] < 1])

    pid = meter.all_pulses[0].unique_id
    assert piece.pulse_from_id(pid) == meter.all_pulses[0]


def test_meters_and_instrumentation_update_duration_arrays():
    piece = build_simple_piece()
    original = json.dumps(piece.dur_array_grid)

    m = Meter([1], start_time=2.1, tempo=60)
    piece.add_meter(m)
    piece.remove_meter(m)
    assert json.dumps(piece.dur_array_grid) == original

    piece.instrumentation.append(Instrument.Vocal_M)
    new_phrase = Phrase({
        'trajectories': [Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 2})],
        'raga': piece.raga,
    })
    piece.phrase_grid.append([new_phrase])
    piece.dur_array_grid.append([])
    piece.section_starts_grid.append([0])
    piece.section_cat_grid.append([init_sec_categorization() for _ in piece.section_cat_grid[0]])
    piece.ad_hoc_section_cat_grid.append([[] for _ in piece.ad_hoc_section_cat_grid[0]])
    piece.dur_array_from_phrases()
    assert len(piece.dur_array_grid) == 2

    piece.instrumentation.pop()
    piece.phrase_grid.pop()
    piece.dur_array_grid.pop()
    piece.section_starts_grid.pop()
    piece.section_cat_grid.pop()
    piece.ad_hoc_section_cat_grid.pop()
    piece.dur_array_from_phrases()
    assert json.dumps(piece.dur_array_grid) == original


def test_piece_serialization_reconnects_groups_and_fixes_slide():
    raga = Raga()
    artics = {'0.00': Articulation({'name': 'slide'})}
    t1 = Trajectory({'num': 0, 'pitches': [Pitch()], 'dur_tot': 0.5, 'articulations': artics})
    t2 = Trajectory({'num': 1, 'pitches': [Pitch()], 'dur_tot': 0.5})
    phrase = Phrase({'trajectories': [t1, t2], 'raga': raga})
    group = Group({'trajectories': [t1, t2]})
    phrase.groups_grid[0].append(group)

    piece = Piece({'phrases': [phrase], 'raga': raga, 'instrumentation': [Instrument.Sitar]})
    clone = Piece.from_json(piece.to_json())

    reconstructed = clone.phrases[0].groups_grid[0][0]
    assert isinstance(reconstructed, Group)
    assert reconstructed.trajectories[0] is clone.phrases[0].trajectory_grid[0][0]
    assert reconstructed.trajectories[1] is clone.phrases[0].trajectory_grid[0][1]
    assert clone.phrases[0].trajectory_grid[0][0].articulations['0.00'].name == 'pluck'


# ----------------------------------------------------------------------
# Track Titles Tests (Issue #44)
# ----------------------------------------------------------------------

def test_track_titles_default_initialization():
    """Test that trackTitles defaults to empty strings matching instrumentation length."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    # Single instrument
    piece = Piece({
        'phrases': [phrase],
        'instrumentation': [Instrument.Sitar],
        'raga': raga
    })
    assert piece.track_titles == ['']

    # Multiple instruments
    piece_multi = Piece({
        'phraseGrid': [[phrase], [phrase]],
        'instrumentation': [Instrument.Sitar, Instrument.Vocal_M],
        'raga': raga
    })
    assert piece_multi.track_titles == ['', '']


def test_track_titles_explicit_values():
    """Test trackTitles with explicit values provided."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    piece = Piece({
        'phraseGrid': [[phrase], [phrase], [phrase]],
        'instrumentation': [Instrument.Sarangi, Instrument.Sarangi, Instrument.Sarangi],
        'trackTitles': ['Lead Melody', 'Harmony', 'Drone'],
        'raga': raga
    })
    assert piece.track_titles == ['Lead Melody', 'Harmony', 'Drone']


def test_track_titles_length_synchronization_shorter():
    """Test that shorter trackTitles array is padded with empty strings."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    piece = Piece({
        'phraseGrid': [[phrase], [phrase], [phrase]],
        'instrumentation': [Instrument.Sarangi, Instrument.Sarangi, Instrument.Sarangi],
        'trackTitles': ['Lead'],
        'raga': raga
    })
    assert len(piece.track_titles) == 3
    assert piece.track_titles == ['Lead', '', '']


def test_track_titles_length_synchronization_longer():
    """Test that longer trackTitles array is truncated to match instrumentation."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    piece = Piece({
        'phrases': [phrase],
        'instrumentation': [Instrument.Sitar],
        'trackTitles': ['Main', 'Extra', 'Another'],
        'raga': raga
    })
    assert len(piece.track_titles) == 1
    assert piece.track_titles == ['Main']


def test_track_titles_type_validation_not_list():
    """Test that non-list trackTitles raises TypeError."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    with pytest.raises(TypeError, match="Parameter 'trackTitles' must be a list"):
        Piece({
            'phrases': [phrase],
            'instrumentation': [Instrument.Sitar],
            'trackTitles': 'not a list',
            'raga': raga
        })


def test_track_titles_type_validation_non_string_items():
    """Test that trackTitles with non-string items raises TypeError."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    with pytest.raises(TypeError, match="All items in 'trackTitles' must be strings"):
        Piece({
            'phrases': [phrase],
            'instrumentation': [Instrument.Sitar],
            'trackTitles': [123],
            'raga': raga
        })


def test_track_titles_serialization_round_trip():
    """Test that trackTitles survives serialization and deserialization."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    piece = Piece({
        'phraseGrid': [[phrase], [phrase]],
        'instrumentation': [Instrument.Sitar, Instrument.Sarangi],
        'trackTitles': ['Melody', 'Harmony'],
        'raga': raga
    })

    # Serialize and deserialize
    json_obj = piece.to_json()
    assert 'trackTitles' in json_obj
    assert json_obj['trackTitles'] == ['Melody', 'Harmony']

    copy = Piece.from_json(json_obj)
    assert copy.track_titles == ['Melody', 'Harmony']

    # Round trip again
    assert copy.to_json()['trackTitles'] == piece.to_json()['trackTitles']


def test_track_titles_empty_string_values():
    """Test that empty strings are valid trackTitles values."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    piece = Piece({
        'phraseGrid': [[phrase], [phrase]],
        'instrumentation': [Instrument.Sitar, Instrument.Vocal_M],
        'trackTitles': ['', ''],
        'raga': raga
    })
    assert piece.track_titles == ['', '']


def test_track_titles_sarangi_trio_use_case():
    """Test the sarangi trio use case from the issue."""
    raga = Raga()
    phrase = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    piece = Piece({
        'phraseGrid': [[phrase], [phrase], [phrase]],
        'instrumentation': [Instrument.Sarangi, Instrument.Sarangi, Instrument.Sarangi],
        'trackTitles': ['Lead Melody', 'Harmony', 'Drone'],
        'raga': raga
    })

    assert len(piece.track_titles) == len(piece.instrumentation)
    assert piece.track_titles[0] == 'Lead Melody'
    assert piece.track_titles[1] == 'Harmony'
    assert piece.track_titles[2] == 'Drone'

    # Verify serialization preserves the titles
    json_obj = piece.to_json()
    copy = Piece.from_json(json_obj)
    assert copy.track_titles == piece.track_titles


# ----------------------------------------------------------------------
# is_section_start Migration Tests (Issue #47)
# ----------------------------------------------------------------------

def test_section_starts_grid_migration_to_phrases():
    """Test migration from old sectionStartsGrid to phrase-level is_section_start."""
    raga = Raga()
    phrase1 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    phrase2 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    phrase3 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    # Create piece with old-style sectionStartsGrid
    piece = Piece({
        'phraseGrid': [[phrase1, phrase2, phrase3]],
        'sectionStartsGrid': [[0, 2]],  # First and third phrases are section starts
        'raga': raga,
        'instrumentation': [Instrument.Sitar]
    })

    # Verify migration happened
    assert piece.phrase_grid[0][0].is_section_start is True
    assert piece.phrase_grid[0][1].is_section_start is False
    assert piece.phrase_grid[0][2].is_section_start is True


def test_section_starts_grid_migration_multi_track():
    """Test migration for multi-track pieces."""
    raga = Raga()
    p1 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    p2 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    p3 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})
    p4 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'raga': raga})

    piece = Piece({
        'phraseGrid': [[p1, p2], [p3, p4]],
        'sectionStartsGrid': [[0, 1], [1]],  # Different section starts per track
        'raga': raga,
        'instrumentation': [Instrument.Sitar, Instrument.Vocal_M]
    })

    # Track 0
    assert piece.phrase_grid[0][0].is_section_start is True
    assert piece.phrase_grid[0][1].is_section_start is True

    # Track 1
    assert piece.phrase_grid[1][0].is_section_start is False
    assert piece.phrase_grid[1][1].is_section_start is True


def test_phrases_with_is_section_start_preserved():
    """Test that phrases created with is_section_start keep their values."""
    raga = Raga()
    phrase1 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'is_section_start': True, 'raga': raga})
    phrase2 = Phrase({'trajectories': [Trajectory({'dur_tot': 1})], 'is_section_start': False, 'raga': raga})

    piece = Piece({
        'phraseGrid': [[phrase1, phrase2]],
        'raga': raga,
        'instrumentation': [Instrument.Sitar]
    })

    # Migration should not override existing is_section_start values
    # Since sectionStartsGrid defaults to [[0]], phrase1 should remain True
    assert piece.phrase_grid[0][0].is_section_start is True
    # phrase2 will be set based on sectionStartsGrid (which has 0 but not 1)
    assert piece.phrase_grid[0][1].is_section_start is False



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class MusicalTime:
    """Represents a musical time position within a meter."""
    cycle_number: int
    hierarchical_position: List[int] 
    fractional_beat: float
    
    def __post_init__(self):
        """Validate parameters after initialization."""
        if self.cycle_number < 0:
            raise ValueError("cycle_number must be non-negative")
        if not all(pos >= 0 for pos in self.hierarchical_position):
            raise ValueError("All hierarchical positions must be non-negative")
        if not (0.0 <= self.fractional_beat < 1.0):
            raise ValueError("fractional_beat must be in range [0.0, 1.0)")
    
    def __str__(self) -> str:
        """Compact string representation: C{cycle}:{hierarchy}+{fraction}"""
        hierarchy_str = ".".join(map(str, self.hierarchical_position))
        return f"C{self.cycle_number}:{hierarchy_str}+{self.fractional_beat:.3f}"
    
    def to_readable_string(self) -> str:
        """Human-readable format with level names."""
        if not self.hierarchical_position:
            return f"Cycle {self.cycle_number + 1}"
        
        level_names = ["Beat", "Subdivision", "Sub-subdivision", "Sub-sub-subdivision"]
        
        parts = []
        for i, pos in enumerate(self.hierarchical_position):
            if i < len(level_names):
                name = level_names[i]
            else:
                name = f"Sub^{i-1}-subdivision"
            parts.append(f"{name} {pos + 1}")  # 1-indexed for display
        
        base = f"Cycle {self.cycle_number + 1}: {', '.join(parts)}"
        if self.fractional_beat > 0:
            base += f" + {self.fractional_beat:.3f}"
        return base
    
    @property
    def beat(self) -> int:
        """Beat number (0-indexed)"""
        return self.hierarchical_position[0] if self.hierarchical_position else 0
    
    @property 
    def subdivision(self) -> Optional[int]:
        """Subdivision number (0-indexed), None if hierarchy has only 1 level"""
        return self.hierarchical_position[1] if len(self.hierarchical_position) > 1 else None
        
    @property
    def sub_subdivision(self) -> Optional[int]:
        """Sub-subdivision number (0-indexed), None if hierarchy has < 3 levels"""
        return self.hierarchical_position[2] if len(self.hierarchical_position) > 2 else None
    
    def get_level(self, level: int) -> Optional[int]:
        """Get position at arbitrary hierarchical level (0=beat, 1=subdivision, etc.)"""
        return self.hierarchical_position[level] if level < len(self.hierarchical_position) else None
    
    @property
    def hierarchy_depth(self) -> int:
        """Number of hierarchical levels in this musical time"""
        return len(self.hierarchical_position)


"""
Utility functions for idtap
"""
import humps
from typing import Dict, Any, List, Union


def selective_decamelize(obj: Dict[str, Any], preserve_keys: List[str] = None) -> Dict[str, Any]:
    """
    Decamelize dictionary keys selectively, preserving certain nested structures.
    
    Args:
        obj: Dictionary to decamelize
        preserve_keys: List of top-level keys whose nested content should not be decamelized
        
    Returns:
        Dictionary with camelCase keys converted to snake_case, except for preserved nested structures
    """
    if preserve_keys is None:
        preserve_keys = [
            'categorization_grid', 
            'categorizationGrid',
            'section_categorization',
            'sectionCategorization', 
            'ad_hoc_categorization',
            'adHocCategorization',
            'rule_set',
            'ruleSet',
            'tuning'
        ]
    
    # First convert top-level keys to snake_case
    result = {}
    for key, value in obj.items():
        snake_key = humps.decamelize(key)
        
        # Check if this key should preserve its nested structure
        if key in preserve_keys or snake_key in preserve_keys:
            result[snake_key] = value
        else:
            # For other keys, recursively decamelize if it's a dict
            if isinstance(value, dict):
                result[snake_key] = humps.decamelize(value)
            elif isinstance(value, list) and value and isinstance(value[0], dict):
                result[snake_key] = [humps.decamelize(item) if isinstance(item, dict) else item for item in value]
            else:
                result[snake_key] = value
    
    return result


def to_camel_case(obj: Union[Dict[str, Any], List, Any]) -> Union[Dict[str, Any], List, Any]:
    """
    Convert dictionary keys from snake_case to camelCase.
    
    Args:
        obj: Object to convert (dict, list, or other)
        
    Returns:
        Object with camelCase keys
    """
    return humps.camelize(obj)


def to_snake_case(obj: Union[Dict[str, Any], List, Any]) -> Union[Dict[str, Any], List, Any]:
    """
    Convert dictionary keys from camelCase to snake_case.
    
    Args:
        obj: Object to convert (dict, list, or other)
        
    Returns:
        Object with snake_case keys
    """
    return humps.decamelize(obj)



import os
import sys
import math
import pytest

sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.raga import Raga, yaman_rule_set, et_tuning
from idtap.classes.pitch import Pitch

base_tuning = et_tuning

base_ratios = [
    base_tuning['sa'],
    base_tuning['re']['raised'],
    base_tuning['ga']['raised'],
    base_tuning['ma']['raised'],
    base_tuning['pa'],
    base_tuning['dha']['raised'],
    base_tuning['ni']['raised'],
]

custom_rule_set = {
    'sa': True,
    're': {'lowered': True, 'raised': False},
    'ga': {'lowered': False, 'raised': True},
    'ma': {'lowered': True, 'raised': True},
    'pa': True,
    'dha': {'lowered': True, 'raised': False},
    'ni': {'lowered': False, 'raised': True},
}

ratio_mapping = [
    ('sa', True),
    ('re', False),
    ('ga', True),
    ('ma', False),
    ('ma', True),
    ('pa', True),
    ('dha', False),
    ('ni', True),
]

def compute_expected_pitches(r: Raga, low=100, high=800):
    pitches: list[Pitch] = []
    for swara, raised in ratio_mapping:
        ratio = (r.tuning[swara]['raised' if raised else 'lowered']
                 if isinstance(r.tuning[swara], dict)
                 else r.tuning[swara])
        freq = ratio * r.fundamental
        low_exp = math.ceil(math.log2(low / freq))
        high_exp = math.floor(math.log2(high / freq))
        for i in range(low_exp, high_exp + 1):
            pitches.append(Pitch({'swara': swara, 'oct': i, 'raised': raised,
                                  'fundamental': r.fundamental, 'ratios': r.stratified_ratios}))
    pitches.sort(key=lambda p: p.frequency)
    return [p for p in pitches if low <= p.frequency <= high]


def test_default_raga():
    r = Raga()
    assert isinstance(r, Raga)
    assert r.name == 'Yaman'
    assert r.fundamental == 261.63
    assert r.rule_set == yaman_rule_set
    assert r.tuning == base_tuning
    assert r.ratios == base_ratios
    assert r.sargam_letters == ['S', 'R', 'G', 'M', 'P', 'D', 'N']
    assert r.rule_set_num_pitches == 7

    pitch_nums = list(range(12))
    sargam_letters = ['S', None, 'R', None, 'G', None, 'M', 'P', None, 'D', None, 'N']
    for pn in pitch_nums:
        assert r.pitch_number_to_sargam_letter(pn) == sargam_letters[pn]

    single_oct_pns = [0, 2, 4, 6, 7, 9, 11, 12]
    assert r.get_pitch_numbers(0, 12) == single_oct_pns

    pns = [
        -12, -10, -8, -6, -5, -3, -1,
        0, 2, 4, 6, 7, 9, 11,
        12, 14, 16, 18, 19, 21, 23, 24,
    ]
    assert r.get_pitch_numbers(-12, 24) == pns

    sns = [
        -7, -6, -5, -4, -3, -2, -1,
        0, 1, 2, 3, 4, 5, 6,
        7, 8, 9, 10, 11, 12, 13, 14,
    ]
    throw_pns = [
        -11, -9, -7, -4, -2,
        1, 3, 5, 8, 10,
        13, 15, 17, 20, 22,
    ]

    for idx, pn in enumerate(pns):
        assert r.pitch_number_to_scale_number(pn) == sns[idx]

    for pn in throw_pns:
        with pytest.raises(ValueError):
            r.pitch_number_to_scale_number(pn)

    for idx, sn in enumerate(sns):
        assert r.scale_number_to_pitch_number(sn) == pns[idx]

    s_letters = ['S', 'R', 'G', 'M', 'P', 'D', 'N'] * 3 + ['S']
    for idx, sn in enumerate(sns):
        assert r.scale_number_to_sargam_letter(sn) == s_letters[idx]

    p_swaras = [
        5, 6,
        0, 1, 2, 3, 4, 5, 6,
        0, 1, 2, 3, 4, 5, 6,
        0, 1, 2, 3, 4,
    ]
    p_octs = [
        -2, -2,
        -1, -1, -1, -1, -1, -1, -1,
        0, 0, 0, 0, 0, 0, 0,
        1, 1, 1, 1, 1,
    ]
    pitches = [Pitch({'swara': s, 'oct': p_octs[idx]}) for idx, s in enumerate(p_swaras)]
    assert r.get_pitches() == pitches

    s_ratios = [
        2 ** 0,
        [2 ** (1 / 12), 2 ** (2 / 12)],
        [2 ** (3 / 12), 2 ** (4 / 12)],
        [2 ** (5 / 12), 2 ** (6 / 12)],
        2 ** (7 / 12),
        [2 ** (8 / 12), 2 ** (9 / 12)],
        [2 ** (10 / 12), 2 ** (11 / 12)],
    ]
    assert r.stratified_ratios == s_ratios
    assert r.chikari_pitches == [
        Pitch({'swara': 0, 'oct': 2, 'fundamental': 261.63}),
        Pitch({'swara': 0, 'oct': 1, 'fundamental': 261.63}),
    ]

    hard_coded_freqs = [
        110.00186456141468, 123.47291821345574,
        130.815, 146.83487284959062,
        164.81657214199782, 185.00034716183643,
        196.0010402616231, 220.00372912282936,
        246.94583642691148, 261.63,
        293.66974569918125, 329.63314428399565,
        370.00069432367286, 392.0020805232462,
        440.0074582456587, 493.89167285382297,
        523.26, 587.3394913983625,
        659.2662885679913, 740.0013886473457,
        784.0041610464924,
    ]
    for idx, freq in enumerate(r.get_frequencies()):
        assert math.isclose(pitches[idx].frequency, freq, abs_tol=1e-6)
        assert math.isclose(hard_coded_freqs[idx], freq, abs_tol=1e-6)

    s_names = ['Sa', 'Re', 'Ga', 'Ma', 'Pa', 'Dha', 'Ni']
    assert r.sargam_names == s_names

    json_obj = {
        'name': 'Yaman',
        'fundamental': 261.63,
        'ratios': base_ratios,
        'tuning': base_tuning,
    }
    assert r.to_json() == json_obj


def test_pitch_from_log_freq():
    r = Raga()
    offset = 0.03
    base_log = math.log2(r.fundamental * 2)
    p = r.pitch_from_log_freq(base_log + offset)
    assert isinstance(p, Pitch)
    assert math.isclose(p.frequency, 2 ** (base_log + offset), abs_tol=1e-6)


def test_pitch_string_getters():
    r = Raga()
    pl = r.get_pitches(low=r.fundamental, high=r.fundamental * 1.999)
    solfege = [p.solfege_letter for p in pl]
    pcs = [str(p.chroma) for p in pl]
    western = [p.western_pitch for p in pl]
    assert r.solfege_strings == solfege
    assert r.pc_strings == pcs
    assert r.western_pitch_strings == western


def test_swara_objects():
    r = Raga()
    objs = [
        {'swara': 0, 'raised': True},
        {'swara': 1, 'raised': True},
        {'swara': 2, 'raised': True},
        {'swara': 3, 'raised': True},
        {'swara': 4, 'raised': True},
        {'swara': 5, 'raised': True},
        {'swara': 6, 'raised': True},
    ]
    assert r.swara_objects == objs


def test_ratio_idx_to_tuning_tuple():
    r = Raga()
    mapping = [
        ('sa', None),
        ('re', 'raised'),
        ('ga', 'raised'),
        ('ma', 'raised'),
        ('pa', None),
        ('dha', 'raised'),
        ('ni', 'raised'),
    ]
    for idx, tup in enumerate(mapping):
        assert r.ratio_idx_to_tuning_tuple(idx) == tup


def test_custom_rule_set_pitch_numbers_and_invalid_conversions():
    r = Raga({'rule_set': custom_rule_set, 'fundamental': 200})
    expected = [0, 1, 4, 5, 6, 7, 8, 11]
    assert r.rule_set_num_pitches == len(expected)
    assert r.get_pitch_numbers(0, 11) == expected
    for idx, pn in enumerate(expected):
        assert r.pitch_number_to_scale_number(pn) == idx
    disallowed = [2, 3, 9, 10]
    for pn in disallowed:
        with pytest.raises(ValueError):
            r.pitch_number_to_scale_number(pn)


def test_get_pitches_with_lowered_and_raised_notes():
    r = Raga({'rule_set': custom_rule_set, 'fundamental': 200})
    expected = compute_expected_pitches(r)
    result = r.get_pitches()
    assert len(result) == len(expected)
    for idx, p in enumerate(result):
        assert math.isclose(p.frequency, expected[idx].frequency, abs_tol=1e-6)
        assert p.swara == expected[idx].swara
        assert p.oct == expected[idx].oct
        assert p.raised == expected[idx].raised


def test_pitch_from_log_freq_octave_rounding():
    r = Raga()
    base = math.log2(r.fundamental)
    near = base + 1 - 5e-7
    p1 = r.pitch_from_log_freq(near)
    assert p1.sargam_letter == 'S'
    assert p1.oct == 1
    p2 = r.pitch_from_log_freq(base + 1 + 5e-7)
    assert p2.sargam_letter == 'S'
    assert p2.oct == 1


def test_pitch_from_log_freq_near_exact_octave_offset():
    fundamental = 128.5 - 1e-12
    r = Raga({'fundamental': fundamental})
    log_freq = math.log2(fundamental * 2)
    p = r.pitch_from_log_freq(log_freq)
    assert p.sargam_letter == 'S'
    assert p.oct == 1


def test_ratio_idx_to_tuning_tuple_mixed_rule_set():
    r = Raga({'rule_set': custom_rule_set})
    mapping = [
        ('sa', None),
        ('re', 'lowered'),
        ('ga', 'raised'),
        ('ma', 'lowered'),
        ('ma', 'raised'),
        ('pa', None),
        ('dha', 'lowered'),
        ('ni', 'raised'),
    ]
    for idx, tup in enumerate(mapping):
        assert r.ratio_idx_to_tuning_tuple(idx) == tup


def test_json_round_trip_with_custom_tuning():
    tuning = {
        'sa': 1.01,
        're': {'lowered': 1.02, 'raised': 1.035},
        'ga': {'lowered': 1.04, 'raised': 1.05},
        'ma': {'lowered': 1.06, 'raised': 1.07},
        'pa': 1.08,
        'dha': {'lowered': 1.09, 'raised': 1.1},
        'ni': {'lowered': 1.11, 'raised': 1.12},
    }
    ratios = [
        tuning['sa'],
        tuning['re']['lowered'],
        tuning['ga']['raised'],
        tuning['ma']['lowered'],
        tuning['ma']['raised'],
        tuning['pa'],
        tuning['dha']['lowered'],
        tuning['ni']['raised'],
    ]
    r = Raga({'name': 'Custom', 'fundamental': 300, 'rule_set': custom_rule_set, 'tuning': tuning, 'ratios': ratios})
    json_obj = r.to_json()
    round_trip = Raga.from_json({**json_obj, 'rule_set': custom_rule_set})
    assert round_trip.to_json() == json_obj

# Additional raga utilities
additional_rule_set = {
    'sa': True,
    're': {'lowered': True, 'raised': False},
    'ga': {'lowered': False, 'raised': True},
    'ma': {'lowered': True, 'raised': True},
    'pa': True,
    'dha': {'lowered': False, 'raised': True},
    'ni': {'lowered': True, 'raised': False},
}

fundamental = 200

def et(n: int):
    return 2 ** (n / 12)

expected_ratios = [
    et(0),
    et(1),
    et(4),
    et(5),
    et(6),
    et(7),
    et(9),
    et(10),
]

expected_stratified = [
    et(0),
    [et(1), et(2)],
    [et(3), et(4)],
    [et(5), et(6)],
    et(7),
    [et(8), et(9)],
    [et(10), et(11)],
]

def compute_freqs(r: Raga, low=100, high=800):
    freqs: list[float] = []
    for ratio in expected_ratios:
        base = ratio * r.fundamental
        low_exp = math.ceil(math.log2(low / base))
        high_exp = math.floor(math.log2(high / base))
        for i in range(low_exp, high_exp + 1):
            freqs.append(base * 2 ** i)
    freqs.sort()
    return freqs

mapping_additional = [
    ('sa', None),
    ('re', 'lowered'),
    ('ga', 'raised'),
    ('ma', 'lowered'),
    ('ma', 'raised'),
    ('pa', None),
    ('dha', 'raised'),
    ('ni', 'lowered'),
]

def test_set_ratios_and_stratified_ratios_with_custom_rules():
    r = Raga({'rule_set': additional_rule_set, 'fundamental': fundamental})
    assert r.set_ratios(additional_rule_set) == expected_ratios
    assert len(r.ratios) == len(expected_ratios)
    for idx, ratio in enumerate(expected_ratios):
        assert math.isclose(r.ratios[idx], ratio, abs_tol=1e-6)
    assert len(r.stratified_ratios) == len(expected_stratified)
    for idx, ratio in enumerate(expected_stratified):
        if isinstance(ratio, list):
            assert r.stratified_ratios[idx] == ratio
        else:
            assert math.isclose(r.stratified_ratios[idx], ratio, abs_tol=1e-6)


def test_from_json():
    """Test creating a Raga from JSON data."""
    json_data = {
        "name": "Bhairav",
        "fundamental": 440.0,
        "ruleSet": {
            "sa": True,
            "re": {"lowered": True, "raised": False},
            "ga": {"lowered": False, "raised": True},
            "ma": {"lowered": False, "raised": True},
            "pa": True,
            "dha": {"lowered": True, "raised": False},
            "ni": {"lowered": False, "raised": True}
        }
    }
    
    raga = Raga.from_json(json_data)
    
    assert raga.name == "Bhairav"
    assert raga.fundamental == 440.0
    assert raga.rule_set_num_pitches == 7


def test_rageshree_issue_17():
    """Test for Issue #17: Raga class incorrectly transforms stored ratios.
    
    Rageshree has 6 pitches: S R G m D n (komal Ma, komal Ni, no Pa).
    The stored ratios should be preserved exactly when using from_json().
    
    This test uses a static snapshot of the data structure to ensure
    the Raga class handles it correctly without needing API access.
    """
    # Actual stored data structure from transcription ID: 68bedf76ffd9b2d478ee11f3
    # This represents what comes from the API after get_piece() fetches the ruleSet
    stored_raga_data = {
        'name': 'Rageshree',
        'fundamental': 261.63,
        'ratios': [
            1.0,
            1.122462048309373,
            1.2599210498948732,
            1.3348398541700344,  # komal Ma
            1.681792830507429,   # D
            1.7817974362806785   # komal Ni
        ],
        'ruleSet': {
            'sa': True,
            're': {'lowered': False, 'raised': True},
            'ga': {'lowered': False, 'raised': True},
            'ma': {'lowered': True, 'raised': False},   # Only komal Ma
            'pa': False,                                 # No Pa
            'dha': {'lowered': False, 'raised': True},
            'ni': {'lowered': True, 'raised': False}     # Only komal Ni
        }
    }
    
    # Test loading with from_json (preserve_ratios=True)
    raga = Raga.from_json(stored_raga_data)
    
    # Ratios should be preserved exactly as stored
    assert raga.ratios == stored_raga_data['ratios'], "Ratios should be preserved exactly"
    assert len(raga.ratios) == 6, f"Expected 6 ratios, got {len(raga.ratios)}"
    
    # Verify specific problematic ratios are preserved
    assert abs(raga.ratios[3] - 1.3348398541700344) < 1e-10, "komal Ma ratio should be preserved"
    assert abs(raga.ratios[5] - 1.7817974362806785) < 1e-10, "komal Ni ratio should be preserved"
    
    # Verify no Pa (1.5) is added
    pa_ratio = 1.4983070768766815  # Approximate Pa ratio
    for ratio in raga.ratios:
        assert abs(ratio - pa_ratio) > 0.01, f"Pa (ratio ~{pa_ratio}) should not be present"
    
    # Get pitches and verify correct count in one octave
    pitches = raga.get_pitches(low=261.63, high=523.25)
    assert len(pitches) == 6, f"Expected 6 pitches in octave, got {len(pitches)}"
    
    # Get frequencies and verify correct count
    frequencies = raga.get_frequencies(low=261.63, high=523.25)
    assert len(frequencies) == 6, f"Expected 6 frequencies in octave, got {len(frequencies)}"
    
    # Verify pitch letters are correct (no Pa)
    pitch_letters = [p.sargam_letter for p in pitches]
    assert 'P' not in pitch_letters, "Pa should not be present in pitch letters"
    
    # Verify rule_set has correct number of pitches
    assert raga.rule_set_num_pitches == 6, f"Expected 6 pitches in rule_set, got {raga.rule_set_num_pitches}"


def test_from_json_frequencies_and_helper_mappings():
    r = Raga({'rule_set': additional_rule_set, 'fundamental': fundamental})
    json_obj = r.to_json()
    copy = Raga.from_json({**json_obj, 'rule_set': additional_rule_set})
    assert copy.to_json() == json_obj

    freqs = r.get_frequencies()
    expected_freqs = compute_freqs(r)
    assert len(freqs) == len(expected_freqs)
    for idx, f in enumerate(freqs):
        assert math.isclose(f, expected_freqs[idx], abs_tol=1e-6)

    chosen = freqs[4]
    p = r.pitch_from_log_freq(math.log2(chosen))
    assert isinstance(p, Pitch)
    assert math.isclose(p.frequency, chosen, abs_tol=1e-6)

    for idx, tup in enumerate(mapping_additional):
        assert r.ratio_idx_to_tuning_tuple(idx) == tup


def test_pitch_number_to_scale_number_edge_cases():
    r = Raga({'rule_set': additional_rule_set, 'fundamental': fundamental})
    allowed = r.get_pitch_numbers(0, 11)
    for idx, pn in enumerate(allowed):
        assert r.pitch_number_to_scale_number(pn) == idx
    disallowed = [2, 3, 8, 11]
    for pn in disallowed:
        with pytest.raises(ValueError):
            r.pitch_number_to_scale_number(pn)


def test_raga_conversion_helpers():
    r = Raga({'rule_set': additional_rule_set, 'fundamental': fundamental})
    pitch_numbers = [0, 1, 4, 5, 6, 7, 9, 10]
    letters = ['S', 'r', 'G', 'm', 'M', 'P', 'D', 'n']
    for idx, pn in enumerate(pitch_numbers):
        assert r.pitch_number_to_sargam_letter(pn) == letters[idx]
        assert r.pitch_number_to_scale_number(pn) == idx
        assert r.scale_number_to_pitch_number(idx) == pn
        assert r.scale_number_to_sargam_letter(idx) == letters[idx]
    length = len(pitch_numbers)
    assert r.scale_number_to_pitch_number(length) == pitch_numbers[0] + 12
    assert r.scale_number_to_pitch_number(length + 1) == pitch_numbers[1] + 12
    invalid_pns = [2, 3, 8, 11]
    for pn in invalid_pns:
        assert r.pitch_number_to_sargam_letter(pn) is None
        with pytest.raises(ValueError):
            r.pitch_number_to_scale_number(pn)

# Custom rule set utilities
custom_local_rule_set = {
    'sa': True,
    're': {'lowered': True, 'raised': False},
    'ga': {'lowered': False, 'raised': True},
    'ma': {'lowered': True, 'raised': True},
    'pa': True,
    'dha': {'lowered': True, 'raised': False},
    'ni': {'lowered': False, 'raised': True},
}

expected_ratios_local = [
    2 ** (0 / 12),
    2 ** (1 / 12),
    2 ** (4 / 12),
    2 ** (5 / 12),
    2 ** (6 / 12),
    2 ** (7 / 12),
    2 ** (8 / 12),
    2 ** (11 / 12),
]

ratio_mapping_local = [
    ('sa', True, expected_ratios_local[0]),
    ('re', False, expected_ratios_local[1]),
    ('ga', True, expected_ratios_local[2]),
    ('ma', False, expected_ratios_local[3]),
    ('ma', True, expected_ratios_local[4]),
    ('pa', True, expected_ratios_local[5]),
    ('dha', False, expected_ratios_local[6]),
    ('ni', True, expected_ratios_local[7]),
]

def compute_expected_pitches_local(r: Raga, low=100, high=800):
    pitches: list[Pitch] = []
    for swara, raised, ratio in ratio_mapping_local:
        freq = ratio * r.fundamental
        low_exp = math.ceil(math.log2(low / freq))
        high_exp = math.floor(math.log2(high / freq))
        for i in range(low_exp, high_exp + 1):
            pitches.append(Pitch({'swara': swara, 'oct': i, 'raised': raised,
                                  'fundamental': r.fundamental, 'ratios': r.stratified_ratios}))
    pitches.sort(key=lambda p: p.frequency)
    return [p for p in pitches if low <= p.frequency <= high]

def compute_expected_freqs_local(r: Raga, low=100, high=800):
    freqs: list[float] = []
    for ratio in expected_ratios_local:
        base = ratio * r.fundamental
        low_exp = math.ceil(math.log2(low / base))
        high_exp = math.floor(math.log2(high / base))
        for i in range(low_exp, high_exp + 1):
            freqs.append(base * 2 ** i)
    freqs.sort()
    return freqs

pitch_numbers_single_oct = [0, 1, 4, 5, 6, 7, 8, 11]

def test_custom_rule_set_basic_functions():
    r = Raga({'rule_set': custom_local_rule_set, 'fundamental': fundamental})
    assert r.rule_set_num_pitches == 8
    for idx, pn in enumerate(pitch_numbers_single_oct):
        assert r.get_pitch_numbers(0, 11)[idx] == pn
    for idx, ratio in enumerate(expected_ratios_local):
        assert math.isclose(r.ratios[idx], ratio, abs_tol=1e-6)
    from_set = r.set_ratios(custom_local_rule_set)
    for idx, ratio in enumerate(from_set):
        assert math.isclose(r.ratios[idx], ratio, abs_tol=1e-6)


def test_get_pitches_frequencies_and_flags():
    r = Raga({'rule_set': custom_local_rule_set, 'fundamental': fundamental})
    expected = compute_expected_pitches_local(r)
    result = r.get_pitches()
    assert len(result) == len(expected)
    for idx, p in enumerate(result):
        assert math.isclose(p.frequency, expected[idx].frequency, abs_tol=1e-6)
        assert p.swara == expected[idx].swara
        assert p.oct == expected[idx].oct
        assert p.raised == expected[idx].raised


def test_frequency_helpers_with_custom_rule_set():
    r = Raga({'rule_set': custom_local_rule_set, 'fundamental': fundamental})
    expected_freqs = compute_expected_freqs_local(r)
    freqs = r.get_frequencies()
    assert len(freqs) == len(expected_freqs)
    for idx, f in enumerate(freqs):
        assert math.isclose(f, expected_freqs[idx], abs_tol=1e-6)

    pick_freq = freqs[3]
    p = r.pitch_from_log_freq(math.log2(pick_freq))
    assert math.isclose(p.frequency, pick_freq, abs_tol=1e-6)

    mapping = [
        ('sa', None),
        ('re', 'lowered'),
        ('ga', 'raised'),
        ('ma', 'lowered'),
        ('ma', 'raised'),
        ('pa', None),
        ('dha', 'lowered'),
        ('ni', 'raised'),
    ]
    for idx, tup in enumerate(mapping):
        assert r.ratio_idx_to_tuning_tuple(idx) == tup


def test_model_raga_string_getters():
    r = Raga()
    pl = r.get_pitches(low=r.fundamental, high=r.fundamental * 1.999)
    solfege = [p.solfege_letter for p in pl]
    pcs = [str(p.chroma) for p in pl]
    western = [p.western_pitch for p in pl]
    assert r.solfege_strings == solfege
    assert r.pc_strings == pcs
    assert r.western_pitch_strings == western


# Parameter validation tests
def test_invalid_parameter_rules_instead_of_rule_set():
    """Test that 'rules' parameter raises helpful error message."""
    with pytest.raises(ValueError, match="Parameter 'rules' not supported. Did you mean 'rule_set'?"):
        Raga({'name': 'Bhimpalasi', 'fundamental': 157.6, 'rules': custom_rule_set})


def test_invalid_parameter_multiple_errors():
    """Test multiple invalid parameters with helpful suggestions."""
    with pytest.raises(ValueError) as exc_info:
        Raga({'name': 'Test', 'rules': custom_rule_set, 'fundamental_freq': 200, 'invalid_param': 123})
    
    error_msg = str(exc_info.value)
    assert "Parameter 'rules' not supported. Did you mean 'rule_set'?" in error_msg
    assert "Parameter 'fundamental_freq' not supported. Did you mean 'fundamental'?" in error_msg
    assert "Invalid parameter: 'invalid_param'" in error_msg
    assert "Allowed parameters:" in error_msg


def test_invalid_parameter_raga_name():
    """Test that 'raga_name' parameter raises helpful error message."""
    with pytest.raises(ValueError, match="Parameter 'raga_name' not supported. Did you mean 'name'?"):
        Raga({'raga_name': 'Test', 'fundamental': 200})


def test_invalid_parameter_base_freq():
    """Test that 'base_freq' parameter raises helpful error message."""
    with pytest.raises(ValueError, match="Parameter 'base_freq' not supported. Did you mean 'fundamental'?"):
        Raga({'name': 'Test', 'base_freq': 200})


def test_invalid_parameter_generic():
    """Test generic invalid parameter error."""
    with pytest.raises(ValueError) as exc_info:
        Raga({'name': 'Test', 'unknown_param': 123, 'another_invalid': 456})
    
    error_msg = str(exc_info.value)
    assert "Invalid parameter: 'unknown_param'" in error_msg
    assert "Invalid parameter: 'another_invalid'" in error_msg
    assert "Allowed parameters:" in error_msg


def test_valid_parameters_work():
    """Test that all valid parameters work correctly."""
    # This should not raise any errors
    raga_data = {
        'name': 'Test Raga',
        'fundamental': 200.0,
        'rule_set': custom_rule_set,
        'tuning': et_tuning,
        'ratios': [1.0, 1.1, 1.2]
    }
    r = Raga(raga_data)
    assert r.name == 'Test Raga'
    assert r.fundamental == 200.0


def test_empty_options_dict():
    """Test that empty options dict works (uses defaults)."""
    r = Raga({})
    assert r.name == 'Yaman'
    assert r.fundamental == 261.63


def test_none_options():
    """Test that None options work (uses defaults)."""
    r = Raga(None)
    assert r.name == 'Yaman'
    assert r.fundamental == 261.63


def test_camel_case_conversion():
    """Test that camelCase parameters are converted properly."""
    # This should work because humps.decamelize converts camelCase
    raga_data = {
        'name': 'Test',
        'fundamental': 200.0,
        'ruleSet': custom_rule_set  # camelCase version
    }
    r = Raga(raga_data)
    assert r.name == 'Test'
    assert r.fundamental == 200.0
    assert r.rule_set == custom_rule_set


def test_validation_issue_reproduction():
    """Reproduce the exact issue from the bug report."""
    # This was the problematic code that should now raise an error
    raga_data = {
        'name': 'Bhimpalasi',
        'fundamental': 157.6,
        'rules': {'sa': True, 'ga': {'lowered': True, 'raised': False}}  # Wrong parameter name
    }
    
    with pytest.raises(ValueError, match="Parameter 'rules' not supported. Did you mean 'rule_set'?"):
        Raga(raga_data)


def test_correct_usage_after_fix():
    """Test that the corrected usage works properly."""
    # This is the corrected version that should work (complete rule set)
    raga_data = {
        'name': 'Bhimpalasi',
        'fundamental': 157.6,
        'rule_set': {  # Correct parameter name with complete rule set
            'sa': True,
            're': {'lowered': False, 'raised': True},
            'ga': {'lowered': True, 'raised': False},  # komal ga (characteristic of Bhimpalasi)
            'ma': {'lowered': False, 'raised': True},
            'pa': True,
            'dha': {'lowered': True, 'raised': False},  # komal dha (characteristic of Bhimpalasi)
            'ni': {'lowered': True, 'raised': False}    # komal ni (characteristic of Bhimpalasi)
        }
    }
    
    raga = Raga(raga_data)
    assert raga.name == 'Bhimpalasi'
    assert raga.fundamental == 157.6
    # Test that the rule set was applied correctly for Bhimpalasi characteristics
    assert raga.rule_set['sa'] == True
    assert raga.rule_set['ga']['lowered'] == True   # komal ga
    assert raga.rule_set['ga']['raised'] == False
    assert raga.rule_set['dha']['lowered'] == True  # komal dha  
    assert raga.rule_set['ni']['lowered'] == True   # komal ni


# Comprehensive validation tests for type errors
def test_invalid_type_name():
    """Test that non-string name raises TypeError."""
    with pytest.raises(TypeError, match="Parameter 'name' must be a string"):
        Raga({'name': 123})


def test_invalid_type_fundamental():
    """Test that non-numeric fundamental raises TypeError."""
    with pytest.raises(TypeError, match="Parameter 'fundamental' must be a number"):
        Raga({'fundamental': 'not_a_number'})


def test_invalid_type_rule_set():
    """Test that non-dict rule_set raises TypeError."""
    with pytest.raises(TypeError, match="Parameter 'rule_set' must be a dict"):
        Raga({'rule_set': 'not_a_dict'})


def test_invalid_type_tuning():
    """Test that non-dict tuning raises TypeError."""
    with pytest.raises(TypeError, match="Parameter 'tuning' must be a dict"):
        Raga({'tuning': [1, 2, 3]})


def test_invalid_type_ratios():
    """Test that non-list ratios raises TypeError."""
    with pytest.raises(TypeError, match="Parameter 'ratios' must be a list"):
        Raga({'ratios': 'not_a_list'})


def test_invalid_ratios_non_numeric():
    """Test that non-numeric items in ratios raises TypeError."""
    with pytest.raises(TypeError, match="All items in 'ratios' must be numbers"):
        Raga({'ratios': [1.0, 2.0, 'not_a_number']})


# Value validation tests
def test_negative_fundamental():
    """Test that negative fundamental raises ValueError."""
    with pytest.raises(ValueError, match="Parameter 'fundamental' must be positive"):
        Raga({'fundamental': -100})


def test_zero_fundamental():
    """Test that zero fundamental raises ValueError."""
    with pytest.raises(ValueError, match="Parameter 'fundamental' must be positive"):
        Raga({'fundamental': 0})


def test_fundamental_out_of_range_warning():
    """Test that fundamental frequency outside typical range raises warning."""
    import warnings
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        Raga({'fundamental': 50000})  # Very high frequency
        assert len(w) == 1
        assert "outside typical range" in str(w[0].message)


def test_negative_ratios():
    """Test that negative ratios raise ValueError."""
    with pytest.raises(ValueError, match="All ratios must be positive"):
        Raga({'ratios': [1.0, -2.0, 1.5]})


def test_too_many_ratios():
    """Test that too many ratios raise ValueError."""
    with pytest.raises(ValueError, match="Too many ratios"):
        Raga({'ratios': [1.0] * 13})  # 13 ratios, max is 12


# rule_set structure validation tests
def test_rule_set_missing_swaras():
    """Test that rule_set missing required swaras raises ValueError."""
    incomplete_rule_set = {'sa': True, 're': True}  # Missing other swaras
    with pytest.raises(ValueError, match="rule_set missing required swaras"):
        Raga({'rule_set': incomplete_rule_set})


def test_rule_set_invalid_swaras():
    """Test that rule_set with invalid swaras raises ValueError."""
    invalid_rule_set = {
        'sa': True, 're': True, 'ga': True, 'ma': True, 
        'pa': True, 'dha': True, 'ni': True,
        'invalid_swara': True  # Invalid swara
    }
    with pytest.raises(ValueError, match="rule_set contains invalid swaras"):
        Raga({'rule_set': invalid_rule_set})


def test_rule_set_sa_pa_non_boolean():
    """Test that sa/pa with non-boolean values raise TypeError."""
    invalid_rule_set = {
        'sa': 'not_boolean', 're': True, 'ga': True, 'ma': True,
        'pa': True, 'dha': True, 'ni': True
    }
    with pytest.raises(TypeError, match="rule_set\\['sa'\\] must be boolean"):
        Raga({'rule_set': invalid_rule_set})


def test_rule_set_variable_swara_invalid_type():
    """Test that variable swaras with invalid types raise TypeError."""
    invalid_rule_set = {
        'sa': True, 're': 'invalid_type', 'ga': True, 'ma': True,
        'pa': True, 'dha': True, 'ni': True
    }
    with pytest.raises(TypeError, match="rule_set\\['re'\\] must be boolean or dict"):
        Raga({'rule_set': invalid_rule_set})


def test_rule_set_variable_swara_missing_keys():
    """Test that variable swaras missing lowered/raised keys raise ValueError."""
    invalid_rule_set = {
        'sa': True, 're': {'only_lowered': True}, 'ga': True, 'ma': True,
        'pa': True, 'dha': True, 'ni': True
    }
    with pytest.raises(ValueError, match="rule_set\\['re'\\] missing required keys"):
        Raga({'rule_set': invalid_rule_set})


def test_rule_set_variable_swara_invalid_keys():
    """Test that variable swaras with invalid keys raise ValueError."""
    invalid_rule_set = {
        'sa': True, 
        're': {'lowered': True, 'raised': False, 'invalid_key': True}, 
        'ga': True, 'ma': True, 'pa': True, 'dha': True, 'ni': True
    }
    with pytest.raises(ValueError, match="rule_set\\['re'\\] contains invalid keys"):
        Raga({'rule_set': invalid_rule_set})


def test_rule_set_variable_swara_non_boolean_values():
    """Test that variable swaras with non-boolean values raise TypeError."""
    invalid_rule_set = {
        'sa': True, 
        're': {'lowered': 'not_boolean', 'raised': True}, 
        'ga': True, 'ma': True, 'pa': True, 'dha': True, 'ni': True
    }
    with pytest.raises(TypeError, match="All values in rule_set\\['re'\\] must be boolean"):
        Raga({'rule_set': invalid_rule_set})


# tuning structure validation tests
def test_tuning_missing_swaras():
    """Test that tuning missing required swaras raises ValueError."""
    incomplete_tuning = {'sa': 1.0, 're': 1.1}  # Missing other swaras
    with pytest.raises(ValueError, match="tuning missing required swaras"):
        Raga({'tuning': incomplete_tuning})


def test_tuning_invalid_swaras():
    """Test that tuning with invalid swaras raises ValueError."""
    invalid_tuning = {
        'sa': 1.0, 're': 1.1, 'ga': 1.2, 'ma': 1.3, 
        'pa': 1.5, 'dha': 1.6, 'ni': 1.8,
        'invalid_swara': 2.0  # Invalid swara
    }
    with pytest.raises(ValueError, match="tuning contains invalid swaras"):
        Raga({'tuning': invalid_tuning})


def test_tuning_sa_pa_non_numeric():
    """Test that sa/pa with non-numeric values raise TypeError."""
    invalid_tuning = {
        'sa': 'not_numeric', 're': 1.1, 'ga': 1.2, 'ma': 1.3,
        'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(TypeError, match="tuning\\['sa'\\] must be a number"):
        Raga({'tuning': invalid_tuning})


def test_tuning_sa_pa_negative():
    """Test that sa/pa with negative values raise ValueError."""
    invalid_tuning = {
        'sa': -1.0, 're': 1.1, 'ga': 1.2, 'ma': 1.3,
        'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(ValueError, match="tuning\\['sa'\\] must be positive"):
        Raga({'tuning': invalid_tuning})


def test_tuning_variable_swara_negative_numeric():
    """Test that variable swaras with negative numeric values raise ValueError."""
    invalid_tuning = {
        'sa': 1.0, 're': -1.1, 'ga': 1.2, 'ma': 1.3,
        'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(ValueError, match="tuning\\['re'\\] must be positive"):
        Raga({'tuning': invalid_tuning})


def test_tuning_variable_swara_invalid_type():
    """Test that variable swaras with invalid types raise TypeError."""
    invalid_tuning = {
        'sa': 1.0, 're': 'invalid_type', 'ga': 1.2, 'ma': 1.3,
        'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(TypeError, match="tuning\\['re'\\] must be number or dict"):
        Raga({'tuning': invalid_tuning})


def test_tuning_variable_swara_dict_missing_keys():
    """Test that variable swaras dicts missing keys raise ValueError."""
    invalid_tuning = {
        'sa': 1.0, 're': {'only_lowered': 1.05}, 'ga': 1.2, 'ma': 1.3,
        'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(ValueError, match="tuning\\['re'\\] missing required keys"):
        Raga({'tuning': invalid_tuning})


def test_tuning_variable_swara_dict_invalid_keys():
    """Test that variable swaras dicts with invalid keys raise ValueError."""
    invalid_tuning = {
        'sa': 1.0, 
        're': {'lowered': 1.05, 'raised': 1.1, 'invalid_key': 1.2}, 
        'ga': 1.2, 'ma': 1.3, 'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(ValueError, match="tuning\\['re'\\] contains invalid keys"):
        Raga({'tuning': invalid_tuning})


def test_tuning_variable_swara_dict_non_numeric_values():
    """Test that variable swaras dicts with non-numeric values raise TypeError."""
    invalid_tuning = {
        'sa': 1.0, 
        're': {'lowered': 'not_numeric', 'raised': 1.1}, 
        'ga': 1.2, 'ma': 1.3, 'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(TypeError, match="All values in tuning\\['re'\\] must be numbers"):
        Raga({'tuning': invalid_tuning})


def test_tuning_variable_swara_dict_negative_values():
    """Test that variable swaras dicts with negative values raise ValueError."""
    invalid_tuning = {
        'sa': 1.0, 
        're': {'lowered': -1.05, 'raised': 1.1}, 
        'ga': 1.2, 'ma': 1.3, 'pa': 1.5, 'dha': 1.6, 'ni': 1.8
    }
    with pytest.raises(ValueError, match="All values in tuning\\['re'\\] must be positive"):
        Raga({'tuning': invalid_tuning})


# Test comprehensive validation - everything should work with valid complete data
def test_complete_valid_data():
    """Test that comprehensive valid data works without errors."""
    valid_data = {
        'name': 'Test Raga',
        'fundamental': 220.0,
        'rule_set': {
            'sa': True,
            're': {'lowered': True, 'raised': False},
            'ga': {'lowered': False, 'raised': True},
            'ma': {'lowered': True, 'raised': True},
            'pa': True,
            'dha': {'lowered': False, 'raised': True},
            'ni': {'lowered': True, 'raised': False}
        },
        'tuning': {
            'sa': 1.0,
            're': {'lowered': 1.05, 'raised': 1.1},
            'ga': {'lowered': 1.15, 'raised': 1.2},
            'ma': {'lowered': 1.3, 'raised': 1.35},
            'pa': 1.5,
            'dha': {'lowered': 1.6, 'raised': 1.65},
            'ni': {'lowered': 1.75, 'raised': 1.8}
        },
        'ratios': [1.0, 1.05, 1.2, 1.3, 1.35, 1.5, 1.6, 1.75]
    }
    
    # This should work without any errors
    raga = Raga(valid_data)
    assert raga.name == 'Test Raga'
    assert raga.fundamental == 220.0



import os
import sys
import json
import random
import pytest

sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.automation import Automation

# Tests mirror src/ts/tests/automation.test.ts

def test_automation_basic():
    a = Automation()
    assert isinstance(a, Automation)
    assert a.values[0]['norm_time'] == 0
    assert a.values[0]['value'] == 1
    val_curve = a.generate_value_curve(0.1, 1)
    assert len(val_curve) == 11
    assert val_curve[0] == pytest.approx(1)

    with pytest.raises(SyntaxError):
        a.add_value(1.5, 0.5)
    with pytest.raises(SyntaxError):
        a.add_value(-0.5, 0.5)
    with pytest.raises(SyntaxError):
        a.add_value(0.5, -0.5)
    with pytest.raises(SyntaxError):
        a.add_value(0.5, 1.5)

    a.add_value(1, 0)
    assert len(a.values) == 2
    val_curve2 = a.generate_value_curve(0.1, 1)
    for i in range(11):
        assert val_curve2[i] == pytest.approx(1 - i / 10)

    a.add_value(0.5, 0.2)
    assert len(a.values) == 3
    val_curve3 = a.generate_value_curve(0.1, 1)
    expected_vals = [
        1, 0.84, 0.68, 0.52, 0.36, 0.2,
        0.16, 0.12, 0.08, 0.04, 0
    ]
    for i in range(11):
        assert val_curve3[i] == pytest.approx(expected_vals[i])


def test_partition_and_compress():
    orig = Automation()
    orig.add_value(1, 0)
    dur_array = [0.4, 0.6]
    children = orig.partition(dur_array)
    assert len(children) == 2
    c1, c2 = children
    assert len(c1.values) == 2
    assert len(c2.values) == 2
    assert c1.values[0]['norm_time'] == 0
    assert c1.values[0]['value'] == 1
    assert c1.values[1]['norm_time'] == 1
    assert c1.values[1]['value'] == pytest.approx(0.6)
    assert c2.values[0]['norm_time'] == 0
    assert c2.values[0]['value'] == pytest.approx(0.6)
    assert c2.values[1]['norm_time'] == 1
    assert c2.values[1]['value'] == 0
    assert c1.value_at_x(0.5) == pytest.approx(0.8)
    assert c2.value_at_x(0.5) == pytest.approx(0.3)

    compressed = Automation.compress(children, dur_array)
    assert len(compressed.values) == 2
    assert compressed.values[0]['norm_time'] == 0
    assert compressed.values[0]['value'] == 1
    assert compressed.values[1]['norm_time'] == 1
    assert compressed.values[1]['value'] == 0

    orig2 = Automation()
    times = [0.13, 0.38, 0.44, 0.6, 0.77777, 0.912345]
    vals = [random.random() for _ in times]
    for t, v in zip(times, vals):
        orig2.add_value(t, v)
    dur_array2 = [0.25, 0.3, 0.45]
    children2 = orig2.partition(dur_array2)
    assert len(children2) == 3
    compressed2 = Automation.compress(children2, dur_array2)
    assert len(compressed2.values) == 8
    returned_times = [v['norm_time'] for v in compressed2.values]
    suplemented_times = [0] + times + [1]
    for r, s in zip(returned_times, suplemented_times):
        assert r == pytest.approx(s)

    a1 = Automation()
    a2 = Automation()
    dur_array3 = [0.4, 0.6]
    new_auto = Automation.compress([a1, a2], dur_array3)
    assert len(new_auto.values) == 2


def test_from_json_round_trip():
    orig = Automation()
    orig.add_value(0.3, 0.7)
    orig.add_value(0.8, 0.4)
    json_obj = json.loads(json.dumps({'values': orig.values}))
    clone = Automation.from_json(json_obj)
    assert isinstance(clone, Automation)
    assert clone.values == orig.values


def test_remove_value_bounds():
    a = Automation()
    a.add_value(0.5, 0.5)
    length = len(a.values)
    with pytest.raises(SyntaxError):
        a.remove_value(-1)
    with pytest.raises(SyntaxError):
        a.remove_value(length)
    with pytest.raises(SyntaxError):
        a.remove_value(0)
    with pytest.raises(SyntaxError):
        a.remove_value(length - 1)
    a.remove_value(1)
    assert len(a.values) == length - 1


def test_value_at_x_bounds():
    a = Automation()
    with pytest.raises(SyntaxError):
        a.value_at_x(-0.1)
    with pytest.raises(SyntaxError):
        a.value_at_x(1.1)


def test_partition_with_zero_length_segment():
    orig = Automation()
    orig.add_value(1, 0)
    dur_array = [0.4, 0, 0.6]
    parts = orig.partition(dur_array)
    assert len(parts) == 3

    env1 = parts[0].generate_value_curve(0.1, 1)
    env2 = parts[1].generate_value_curve(0.1, 1)
    env3 = parts[2].generate_value_curve(0.1, 1)

    expected1 = [
        1, 0.96, 0.92, 0.88, 0.84,
        0.8, 0.76, 0.72, 0.68,
        0.64, 0.6
    ]
    for i in range(len(env1)):
        assert env1[i] == pytest.approx(expected1[i])

    for val in env2:
        assert val == pytest.approx(0.6)

    expected3 = [0.6, 0.54, 0.48, 0.42, 0.36, 0.3, 0.24, 0.18, 0.12, 0.06, 0]
    for i in range(len(env3)):
        assert env3[i] == pytest.approx(expected3[i])


def test_value_at_x_before_first_value():
    auto = Automation({
        'values': [
            {'norm_time': 0.2, 'value': 0.5},
            {'norm_time': 1, 'value': 1}
        ]
    })
    with pytest.raises(SyntaxError):
        auto.value_at_x(0.1)




"""HTTP client for the Swara Studio API."""

from __future__ import annotations

from typing import Any, Dict, Optional, Union, List, Callable

import json
from pathlib import Path

import requests
import os

from idtap.classes.piece import Piece

from .auth import login_google, load_token
from .secure_storage import SecureTokenStorage
from .query import Query
from .query_types import (
    QueryType, MultipleReturnType, CategoryType, 
    DesignatorType, SegmentationType, QueryAnswerType
)
from .classes.pitch import Pitch


class SwaraClient:
    """Minimal client wrapping the public API served at https://swara.studio."""

    def __init__(
        self,
        base_url: str = "https://swara.studio/",
        token_path: str | Path | None = None,
        auto_login: bool = True,
    ) -> None:
        self.base_url = base_url.rstrip("/") + "/"
        
        # Initialize secure storage
        self.secure_storage = SecureTokenStorage()
        
        # Keep token_path for backwards compatibility
        self.token_path = Path(token_path or os.environ.get("SWARA_TOKEN_PATH", "~/.swara/token.json")).expanduser() if token_path else None
        
        self.auto_login = auto_login
        self.token: Optional[str] = None
        self.user: Optional[Dict[str, Any]] = None
        self.load_token()
        
        if self.token is None and self.auto_login:
            try:
                login_google(base_url=self.base_url, storage=self.secure_storage)
                self.load_token()
            except Exception as e:
                print(f"Failed to log in to Swara Studio: {e}")
                raise
                
    @property
    def user_id(self) -> Optional[str]:
        """Return the user ID if available, otherwise ``None``."""
        if self.user:
            return self.user.get("_id") or self.user.get("sub")
        return None

    # ---- auth utilities ----
    def load_token(self, token_path: Optional[str | Path] = None) -> None:
        """Load saved token and profile information from secure storage."""
        try:
            # Use the new secure storage with backwards compatibility
            legacy_path = Path(token_path or self.token_path) if (token_path or self.token_path) else None
            data = load_token(storage=self.secure_storage, token_path=legacy_path)
            
            if data:
                # Check if tokens are expired and need refresh
                if self.secure_storage.is_token_expired(data):
                    print("⚠️  Stored tokens are expired. Please re-authenticate.")
                    # Clear expired tokens
                    self.secure_storage.clear_tokens()
                    self.token = None
                    self.user = None
                    return
                
                self.token = data.get("id_token") or data.get("token")
                self.user = data.get("profile") or data.get("user")
            else:
                self.token = None
                self.user = None
        except Exception as e:
            print(f"Failed to load tokens: {e}")
            self.token = None
            self.user = None

    def get_auth_info(self) -> Dict[str, Any]:
        """Get information about the current authentication and storage setup.
        
        Returns:
            Dict containing authentication status and storage information
        """
        storage_info = self.secure_storage.get_storage_info()
        return {
            "authenticated": self.token is not None,
            "user_id": self.user_id,
            "user_email": self.user.get("email") if self.user else None,
            "storage_info": storage_info,
            "token_expired": False if not self.token else None
        }

    def _auth_headers(self) -> Dict[str, str]:
        if self.token:
            return {"Authorization": f"Bearer {self.token}"}
        return {}

    def _post_json(self, endpoint: str, payload: Dict[str, Any]) -> Any:
        url = self.base_url + endpoint
        headers = self._auth_headers()
        response = requests.post(url, json=payload, headers=headers, timeout=1800)  # 30 minutes
        response.raise_for_status()
        if response.content:
            return response.json()
        return None

    def _get(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Any:
        url = self.base_url + endpoint
        headers = self._auth_headers()
        response = requests.get(url, params=params, headers=headers, timeout=1800)  # 30 minutes
        response.raise_for_status()
        ctype = response.headers.get("Content-Type", "")
        if ctype.startswith("application/json"):
            return response.json()
        return response.content

    # ---- API methods ----
    def get_piece(self, piece_id: str, fetch_rule_set: bool = True) -> Any:
        """Return transcription JSON for the given id.
        
        Args:
            piece_id: The ID of the piece to fetch
            fetch_rule_set: If True and raga has no ruleSet, fetch it from database
            
        Returns:
            Dictionary containing the piece data with ruleSet populated if needed
        """
        # Check waiver and prompt if needed
        self._prompt_for_waiver_if_needed()
        piece_data = self._get(f"api/transcription/{piece_id}")
        
        # If fetch_rule_set is True and there's a raga without a ruleSet, fetch it
        if fetch_rule_set and 'raga' in piece_data:
            raga_data = piece_data['raga']
            if 'ruleSet' not in raga_data or not raga_data.get('ruleSet'):
                raga_name = raga_data.get('name')
                if raga_name and raga_name != 'Yaman':
                    try:
                        # Fetch the rule_set from the database
                        raga_rules = self.get_raga_rules(raga_name)
                        if 'rules' in raga_rules:
                            piece_data['raga']['ruleSet'] = raga_rules['rules']
                    except:
                        # If fetch fails, leave it as is
                        pass
        
        return piece_data

    def excel_data(self, piece_id: str) -> bytes:
        """Export transcription data as Excel file."""
        # Check waiver and prompt if needed
        self._prompt_for_waiver_if_needed()
        return self._get(f"api/transcription/{piece_id}/excel")

    def json_data(self, piece_id: str) -> bytes:
        """Export transcription data as JSON file."""
        # Check waiver and prompt if needed
        self._prompt_for_waiver_if_needed()
        return self._get(f"api/transcription/{piece_id}/json")

    def save_piece(self, piece: Dict[str, Any]) -> Any:
        """Save transcription using authenticated API route."""
        return self._post_json("api/transcription", piece)

    def insert_new_transcription(self, piece: Dict[str, Any]) -> Any:
        """Insert a new transcription document as the current authenticated user."""
        if not self.user_id:
            raise RuntimeError("Not authenticated: cannot insert new transcription")
        payload = dict(piece)
        payload["userID"] = self.user_id
        return self._post_json("insertNewTranscription", payload)

    def _prompt_for_waiver_if_needed(self) -> None:
        """Interactively prompt user to agree to waiver if not already agreed."""
        if self.has_agreed_to_waiver():
            return
            
        print("\n" + "=" * 60)
        print("📋 IDTAP RESEARCH WAIVER REQUIRED")
        print("=" * 60)
        print("\nBefore accessing transcription data, you must agree to the following terms:\n")
        
        waiver_text = self.get_waiver_text()
        print(waiver_text)
        
        print("\n" + "=" * 60)
        
        while True:
            response = input("Do you agree to these terms? (yes/no): ").strip().lower()
            
            if response == "yes":
                print("\nSubmitting waiver agreement...")
                try:
                    self.agree_to_waiver(i_agree=True)
                    print("✅ Waiver agreement successful! You now have access to transcription data.\n")
                    break
                except Exception as e:
                    print(f"❌ Error submitting waiver agreement: {e}")
                    raise
            elif response == "no":
                print("\n👋 You must agree to the waiver to access transcription data.")
                raise RuntimeError("Waiver agreement required but declined by user.")
            else:
                print("Please respond with 'yes' or 'no'.")

    def get_viewable_transcriptions(
        self,
        sort_key: str = "title",
        sort_dir: str | int = 1,
        new_permissions: Optional[bool] = None,
    ) -> Any:
        """Return transcriptions viewable by the user."""
        # Check waiver and prompt if needed
        self._prompt_for_waiver_if_needed()
            
        params = {
            "sortKey": sort_key,
            "sortDir": sort_dir,
            "newPermissions": new_permissions,
        }
        # remove None values
        params = {k: str(v) for k, v in params.items() if v is not None}
        return self._get("api/transcriptions", params=params)


    def update_visibility(
        self,
        artifact_type: str,
        _id: str,
        explicit_permissions: Dict[str, Any],
    ) -> Any:
        payload = {
            "artifactType": artifact_type,
            "_id": _id,
            "explicitPermissions": explicit_permissions,
        }
        return self._post_json("api/visibility", payload)

    def has_agreed_to_waiver(self) -> bool:
        """Check if the current user has agreed to the research waiver.
        
        This makes a fresh API call to get the latest waiver status from the database.
        
        Returns:
            True if user has agreed to waiver, False otherwise
        """
        try:
            # Make a fresh API call to get current user data from database
            fresh_user_data = self._get("api/user")
            return fresh_user_data.get("waiverAgreed", False)
        except Exception:
            # Fall back to cached data if API call fails
            if not self.user:
                return False
            return self.user.get("waiverAgreed", False)

    def get_waiver_text(self) -> str:
        """Get the research waiver text that users must agree to.
        
        Returns:
            The full waiver text
        """
        return ("I agree to only use the IDTAP for scholarly and/or pedagogical purposes. "
                "I understand that any copyrighted materials that I upload to the IDTAP "
                "are liable to be taken down in response to a DMCA takedown notice.")

    def agree_to_waiver(self, i_agree: bool = False) -> Any:
        """Agree to the research waiver after reading it.
        
        You must first read the waiver text using get_waiver_text() and then
        explicitly set i_agree=True to confirm agreement.
        
        Args:
            i_agree: Must be True to confirm you have read and agree to the waiver
        
        Returns:
            Server response confirming waiver agreement
            
        Raises:
            RuntimeError: If not authenticated or if i_agree is not True
        """
        if not self.user_id:
            raise RuntimeError("Not authenticated: cannot agree to waiver")
        
        if not i_agree:
            waiver_text = self.get_waiver_text()
            raise RuntimeError(
                f"You must read and agree to the research waiver before accessing transcriptions.\n\n"
                f"WAIVER TEXT:\n{waiver_text}\n\n"
                f"If you agree to these terms, call: client.agree_to_waiver(i_agree=True)"
            )
            
        payload = {"userID": self.user_id}
        result = self._post_json("api/agreeToWaiver", payload)
        
        # Update local user object to reflect waiver agreement
        if self.user:
            self.user["waiverAgreed"] = True
        
        return result

    def upload_audio(
        self,
        file_path: str,
        metadata: "AudioMetadata",
        audio_event: Optional["AudioEventConfig"] = None,
        progress_callback: Optional[Callable[[float], None]] = None
    ) -> "AudioUploadResult":
        """Upload audio recording with comprehensive metadata.

        Requires the `requests-toolbelt` library for multipart encoding.
        
        Args:
            file_path: Path to the audio file to upload
            metadata: AudioMetadata object with recording information.
                     Ragas can be specified in multiple formats:
                     - AudioRaga objects: AudioRaga(name="Rageshree") (recommended)
                     - Strings: "Rageshree" (auto-converted to AudioRaga)
                     - Name dicts: {"name": "Rageshree"} (auto-converted to AudioRaga)
                     - Legacy format: {"Rageshree": {"performance_sections": {}}} (auto-converted)
            audio_event: Optional AudioEventConfig for associating with audio events
            progress_callback: Optional callback for upload progress (0-100)
            
        Returns:
            AudioUploadResult with upload status and file information
            
        Raises:
            FileNotFoundError: If the audio file doesn't exist
            ValueError: If the file is not a supported audio format or metadata validation fails
            RuntimeError: If upload fails
        """
        import os
        from pathlib import Path
        
        # Validate file exists
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"Audio file not found: {file_path}")
        
        file_path_obj = Path(file_path)
        
        # Check file extension
        supported_extensions = {'.mp3', '.wav', '.m4a', '.flac', '.opus', '.ogg'}
        if file_path_obj.suffix.lower() not in supported_extensions:
            raise ValueError(f"Unsupported audio format: {file_path_obj.suffix}. "
                           f"Supported formats: {', '.join(supported_extensions)}")
        
        # Validate metadata early to provide clear error messages
        try:
            # This will trigger raga normalization and validation
            metadata.to_json()
        except ValueError as e:
            raise ValueError(f"Metadata validation failed: {e}")
        
        # Prepare form data
        try:
            # Prepare data fields
            data = {
                'metadata': json.dumps(metadata.to_json()),
            }
            
            if audio_event:
                data['audioEventConfig'] = json.dumps(audio_event.to_json())
            
            # Open file and make request
            with open(file_path, 'rb') as f:
                files = {'audioFile': (file_path_obj.name, f, self._get_mimetype(file_path_obj.suffix))}
                
                from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor

                def progress_monitor(monitor):
                    progress_callback((monitor.bytes_read / monitor.len) * 100)

                fields = {
                    'metadata': json.dumps(metadata.to_json()),
                }
                if audio_event:
                    fields['audioEventConfig'] = json.dumps(audio_event.to_json())
                fields['audioFile'] = (
                    file_path_obj.name,
                    f,
                    self._get_mimetype(file_path_obj.suffix)
                )

                encoder = MultipartEncoder(fields=fields)
                payload = MultipartEncoderMonitor(encoder, progress_monitor) if progress_callback else encoder
                headers = {
                    **self._auth_headers(),
                    'Content-Type': payload.content_type,
                    'Content-Length': str(payload.len),
                }

                response = requests.post(
                    f"{self.base_url}api/uploadAudio",
                    data=payload,
                    headers=headers,
                    timeout=1800
                )
                response.raise_for_status()
                result_data = response.json()
                from .audio_models import AudioUploadResult
                return AudioUploadResult.from_api_response(result_data)
                
        except requests.exceptions.RequestException as e:
            raise RuntimeError(f"Upload failed: {e}")
        except Exception as e:
            raise RuntimeError(f"Upload error: {e}")

    def _get_mimetype(self, extension: str) -> str:
        """Get MIME type for file extension."""
        mime_types = {
            '.mp3': 'audio/mpeg',
            '.wav': 'audio/wav',
            '.m4a': 'audio/m4a',
            '.flac': 'audio/flac',
            '.opus': 'audio/opus',
            '.ogg': 'audio/ogg'
        }
        return mime_types.get(extension.lower(), 'audio/mpeg')

    def get_available_musicians(self) -> List[Dict[str, Any]]:
        """Get list of musicians in database."""
        return self._get("api/musicians")

    def get_available_ragas(self) -> List[str]:
        """Get list of ragas in database."""
        return self._get("api/ragas")
        
    def get_raga_rules(self, raga_name: str) -> Dict[str, Any]:
        """Get pitch rules for a specific raga.
        
        Args:
            raga_name: Name of the raga to get rules for
            
        Returns:
            Dictionary containing the raga's pitch rules and updated date
            
        Raises:
            ValueError: If raga_name is empty or None
            requests.HTTPError: If raga not found or API error
        """
        if not raga_name:
            raise ValueError("Raga name cannot be empty")
            
        params = {"name": raga_name}
        return self._get("api/ragaRules", params)
        
    def get_available_instruments(self, melody_only: bool = False) -> List[str]:
        """Get list of instruments in database."""
        params = {'melody': 'true'} if melody_only else {}
        return self._get("api/instruments", params)
        
    def get_location_hierarchy(self) -> "LocationHierarchy":
        """Get continent/country/city structure."""
        data = self._get("api/locations")
        from .audio_models import LocationHierarchy
        return LocationHierarchy.from_api_response(data)
        
    def get_available_gharanas(self) -> List[Dict[str, Any]]:
        """Get list of gharanas in database."""
        return self._get("api/gharanas")
        
    def get_performance_sections(self) -> List[str]:
        """Get list of performance sections."""
        return self._get("api/performanceSections")
        
    def get_event_types(self) -> List[str]:
        """Get list of audio event types."""
        return self._get("api/eventTypes")
        
    def get_editable_audio_events(self) -> List[Dict[str, Any]]:
        """Get audio events the user can edit."""
        return self._get("api/audioEvents")

    def validate_metadata(self, metadata: "AudioMetadata") -> "ValidationResult":
        """Validate metadata against platform requirements."""
        from .audio_models import ValidationResult
        
        errors = []
        warnings = []
        
        # Get available data for validation
        try:
            available_musicians = [m.get('Full Name', '') for m in self.get_available_musicians()]
            available_ragas = self.get_available_ragas()
            available_instruments = self.get_available_instruments()
            location_hierarchy = self.get_location_hierarchy()
        except Exception as e:
            warnings.append(f"Could not fetch validation data: {e}")
            return ValidationResult(is_valid=True, warnings=warnings)
        
        # Validate musicians
        for musician in metadata.musicians:
            if not musician.name:
                errors.append("Musician name cannot be empty")
            elif musician.name not in available_musicians and musician.name != "Other":
                warnings.append(f"Musician '{musician.name}' not found in database")
            
            if musician.instrument not in available_instruments:
                warnings.append(f"Instrument '{musician.instrument}' not found in database")
        
        # Validate ragas
        for raga in metadata.ragas:
            if not raga.name:
                errors.append("Raga name cannot be empty")
            elif raga.name not in available_ragas:
                warnings.append(f"Raga '{raga.name}' not found in database")
        
        # Validate location
        if metadata.location:
            continents = location_hierarchy.get_continents()
            if metadata.location.continent not in continents:
                warnings.append(f"Continent '{metadata.location.continent}' not found in database")
            else:
                countries = location_hierarchy.get_countries(metadata.location.continent)
                if metadata.location.country not in countries:
                    warnings.append(f"Country '{metadata.location.country}' not found in database")
                elif metadata.location.city:
                    cities = location_hierarchy.get_cities(metadata.location.continent, metadata.location.country)
                    if cities and metadata.location.city not in cities:
                        warnings.append(f"City '{metadata.location.city}' not found in database")
        
        return ValidationResult(is_valid=len(errors) == 0, errors=errors, warnings=warnings)

    def logout(self, confirm: bool = False) -> bool:
        """Log out the current user and clear all stored authentication tokens.
        
        This will:
        - Clear tokens from OS keyring, encrypted storage, and plaintext files
        - Reset the client's authentication state
        - Require re-authentication for future API calls
        
        Args:
            confirm: Set to True to confirm logout without interactive prompt
            
        Returns:
            True if logout was successful, False otherwise
        """
        if not confirm:
            print("🚪 Logging out will clear all stored authentication tokens.")
            print("You will need to re-authenticate to use the API again.")
            user_input = input("Are you sure you want to log out? (yes/no): ").strip().lower()
            if user_input != 'yes':
                print("Logout cancelled.")
                return False
        
        try:
            # Clear tokens from all storage backends
            success = self.secure_storage.clear_tokens()
            
            if success:
                # Reset client authentication state
                self.token = None
                self.user = None
                print("✅ Successfully logged out. All authentication tokens have been cleared.")
                return True
            else:
                print("⚠️  Logout partially successful - some tokens may not have been cleared.")
                return False
                
        except Exception as e:
            print(f"❌ Error during logout: {e}")
            return False

    def download_audio(self, audio_id: str, format: str = "wav") -> bytes:
        """Download audio recording by audio ID.
        
        Args:
            audio_id: The audio recording ID
            format: Audio format (wav, mp3, opus)
            
        Returns:
            Raw audio data as bytes
        """
        if format not in ["wav", "mp3", "opus"]:
            raise ValueError(f"Unsupported audio format: {format}. Use 'wav', 'mp3', or 'opus'")
        
        endpoint = f"audio/{format}/{audio_id}.{format}"
        return self._get(endpoint)

    def download_transcription_audio(self, piece: Union[Dict[str, Any], Piece], format: str = "wav") -> Optional[bytes]:
        """Download audio recording associated with a transcription.
        
        Args:
            piece: Transcription piece data (dict or Piece object)
            format: Audio format (wav, mp3, opus)
            
        Returns:
            Raw audio data as bytes, or None if no audio is associated
        """
        # Extract audio ID from piece
        if hasattr(piece, 'audio_id'):
            audio_id = piece.audio_id
        elif isinstance(piece, dict):
            audio_id = piece.get('audioID')
        else:
            raise TypeError(f"Expected Piece object or dict, got {type(piece)}")
        
        if not audio_id:
            return None
            
        return self.download_audio(audio_id, format)

    def save_audio_file(self, audio_data: bytes, filename: str, filepath: Optional[str] = None) -> str:
        """Save audio data to a file.
        
        Args:
            audio_data: Raw audio data from download_audio()
            filename: Output filename (should include extension)
            filepath: Directory to save file (defaults to user's Downloads folder)
            
        Returns:
            Full path to the saved file
        """
        import os
        from pathlib import Path
        
        if filepath is None:
            # Cross-platform default to Downloads folder
            if os.name == 'nt':  # Windows
                downloads_dir = Path.home() / 'Downloads'
            else:  # macOS, Linux, Unix
                downloads_dir = Path.home() / 'Downloads'
            filepath = str(downloads_dir)
        
        # Ensure directory exists
        Path(filepath).mkdir(parents=True, exist_ok=True)
        
        # Combine path and filename
        full_path = Path(filepath) / filename
        
        with open(full_path, 'wb') as f:
            f.write(audio_data)
            
        return str(full_path)

    def download_and_save_transcription_audio(self, piece: Union[Dict[str, Any], Piece], 
                                              format: str = "wav", 
                                              filepath: Optional[str] = None,
                                              filename: Optional[str] = None) -> Optional[str]:
        """Download and save audio recording associated with a transcription.
        
        Args:
            piece: Transcription piece data (dict or Piece object)
            format: Audio format (wav, mp3, opus)
            filepath: Directory to save file (defaults to Downloads folder)
            filename: Custom filename (defaults to transcription title + ID)
            
        Returns:
            Full path to saved file, or None if no audio is associated
        """
        # Download audio data
        audio_data = self.download_transcription_audio(piece, format)
        if not audio_data:
            return None
        
        # Generate filename if not provided
        if filename is None:
            if hasattr(piece, 'title') and hasattr(piece, '_id'):
                title = piece.title
                piece_id = piece._id
            elif isinstance(piece, dict):
                title = piece.get('title', 'untitled')
                piece_id = piece.get('_id', 'unknown')
            else:
                title = 'untitled'
                piece_id = 'unknown'
            
            # Clean title for filename
            clean_title = "".join(c for c in title if c.isalnum() or c in (' ', '-', '_')).strip()
            filename = f"{clean_title}_{piece_id}.{format}"
        
        # Save file and return path
        return self.save_audio_file(audio_data, filename, filepath)

    def download_spectrogram_data(self, audio_id: str) -> bytes:
        """Download gzip-compressed spectrogram data.

        Args:
            audio_id: The audio recording ID

        Returns:
            Gzipped binary data containing uint8 spectrogram array
        """
        endpoint = f"spec_data/{audio_id}/spec_data.gz"
        return self._get(endpoint)

    def download_spectrogram_metadata(self, audio_id: str) -> Dict[str, Any]:
        """Download spectrogram shape metadata.

        Args:
            audio_id: The audio recording ID

        Returns:
            Dictionary with 'shape' key: [freq_bins, time_frames]
        """
        endpoint = f"spec_data/{audio_id}/spec_shape.json"
        return self._get(endpoint)

    def get_audio_recording(self, audio_id: str) -> Dict[str, Any]:
        """Get audio recording metadata by ID.

        Fetches complete recording metadata including duration, musicians,
        ragas, location, and permissions.

        Args:
            audio_id: The audio recording ID

        Returns:
            Dictionary with recording metadata including:
                - duration: Audio duration in seconds (float)
                - musicians: Dictionary of performer information
                - raags: Dictionary of raga information
                - title: Recording title
                - etc.

        Raises:
            requests.HTTPError: If recording not found (404)
        """
        return self._get("getAudioRecording", params={"_id": audio_id})

    def save_transcription(self, piece: Piece, fill_duration: bool = True) -> Any:
        """Save a transcription piece to the server.
        
        Handles both new transcriptions (without _id) and existing transcriptions (with _id).
        
        Args:
            piece: The Piece object or dict to save
            fill_duration: Whether to automatically fill remaining duration with silence
            
        Returns:
            Server response from the save operation
        """
        # Convert Piece object to dict if needed
        if hasattr(piece, 'to_json'):
            payload = piece.to_json()
        elif isinstance(piece, dict):
            payload = dict(piece)
        else:
            raise TypeError(f"Expected Piece object with to_json() method or dict, got {type(piece)}")
        
        # Fill remaining duration with silence if requested
        if fill_duration and hasattr(piece, 'fill_remaining_duration') and hasattr(piece, 'dur_tot'):
            piece.fill_remaining_duration(piece.dur_tot)
            payload = piece.to_json()
        
        # Set transcriber information from authenticated user if not already set
        if hasattr(piece, 'given_name') and self.user:
            if not getattr(piece, 'given_name', None):
                piece.given_name = self.user.get("given_name", "")
            if not getattr(piece, 'family_name', None):
                piece.family_name = self.user.get("family_name", "")
            if not getattr(piece, 'name', None):
                piece.name = self.user.get("name", "")
        
        # Set default soloist and instrument information if not already set
        if hasattr(piece, 'soloist') and not getattr(piece, 'soloist', None):
            piece.soloist = None
        if hasattr(piece, 'solo_instrument') and not getattr(piece, 'solo_instrument', None):
            instrumentation = getattr(piece, 'instrumentation', [])
            piece.solo_instrument = instrumentation[0] if instrumentation else "Unknown Instrument"
        
        # Regenerate payload after setting user info
        if hasattr(piece, 'to_json'):
            payload = piece.to_json()
        else:
            payload = dict(piece)
        
        # Determine if this is a new or existing transcription
        has_id = payload.get("_id") is not None
        
        if has_id:
            # Existing transcription - use save_piece
            print(f"Updating existing transcription: {payload.get('title', 'untitled')}")
            try:
                response = self.save_piece(payload)
                print("✅ Updated transcription:", response)
                return response
            except Exception as e:
                print("❌ Failed to update transcription:", e)
                raise
        else:
            # New transcription - remove any null _id and use insert_new_transcription
            payload.pop("_id", None)
            print(f"Inserting new transcription: {payload.get('title', 'untitled')}")
            try:
                response = self.insert_new_transcription(payload)
                print("✅ Inserted transcription:", response)
                return response
            except Exception as e:
                print("❌ Failed to insert transcription:", e)
                raise
    
    # ---- Query methods ----
    
    def single_query(
        self,
        transcription_id: str,
        segmentation: Union[SegmentationType, str] = SegmentationType.PHRASE,
        designator: Union[DesignatorType, str] = DesignatorType.INCLUDES,
        category: Union[CategoryType, str] = CategoryType.TRAJECTORY_ID,
        pitch: Optional[Pitch] = None,
        sequence_length: Optional[int] = None,
        trajectory_id: Optional[int] = None,
        vowel: Optional[str] = None,
        consonant: Optional[str] = None,
        instrument_idx: int = 0,
        **kwargs
    ) -> Query:
        """Create and execute a single query on a transcription.
        
        Args:
            transcription_id: ID of the transcription to query
            segmentation: Type of segmentation (phrase, group, etc.)
            designator: Query designator (includes, excludes, etc.)
            category: Query category (trajectoryID, pitch, etc.)
            pitch: Pitch object to search for (if category is pitch)
            sequence_length: Length of trajectory sequences (if needed)
            trajectory_id: Trajectory ID to search for (if category is trajectoryID)
            vowel: Vowel to search for (if category is vowel)
            consonant: Consonant to search for (if category is consonant)
            instrument_idx: Index of instrument track to query
            **kwargs: Additional query parameters
            
        Returns:
            Query object with results
        """
        # Check waiver and prompt if needed
        self._prompt_for_waiver_if_needed()
        
        # Fetch the piece data
        piece_data = self.get_piece(transcription_id)
        piece = Piece.from_json(piece_data)
        
        # Convert string enums to enum objects if needed
        if isinstance(segmentation, str):
            segmentation = SegmentationType(segmentation)
        if isinstance(designator, str):
            designator = DesignatorType(designator)
        if isinstance(category, str):
            category = CategoryType(category)
        
        # Build query options
        query_options = {
            "segmentation": segmentation,
            "designator": designator,
            "category": category,
            "pitch": pitch,
            "sequence_length": sequence_length,
            "trajectory_id": trajectory_id,
            "vowel": vowel,
            "consonant": consonant,
            "instrument_idx": instrument_idx,
            **kwargs
        }
        
        return Query(piece, query_options)
    
    def multiple_query(
        self,
        queries: List[Union[QueryType, Dict[str, Any]]],
        transcription_id: str,
        segmentation: Union[SegmentationType, str] = SegmentationType.PHRASE,
        sequence_length: Optional[int] = None,
        min_dur: float = 0.0,
        max_dur: float = 60.0,
        every: bool = True,
        instrument_idx: int = 0,
    ) -> MultipleReturnType:
        """Execute multiple queries on a transcription and combine results.
        
        Args:
            queries: List of query specifications
            transcription_id: ID of transcription to query
            segmentation: Segmentation type for all queries
            sequence_length: Sequence length for trajectory sequences
            min_dur: Minimum duration filter
            max_dur: Maximum duration filter
            every: If True, require all queries to match; if False, any query can match
            instrument_idx: Index of instrument track to query
            
        Returns:
            Tuple of (trajectories, identifiers, query_answers)
        """
        # Check waiver and prompt if needed
        self._prompt_for_waiver_if_needed()
        
        if not queries:
            raise ValueError("No queries provided")
        
        # Fetch the piece data
        piece_data = self.get_piece(transcription_id)
        piece = Piece.from_json(piece_data)
        
        # Convert string enum to enum object if needed
        if isinstance(segmentation, str):
            segmentation = SegmentationType(segmentation)
        
        # Execute multiple query logic (similar to the static method but integrated)
        output_trajectories: List[List["Trajectory"]] = []
        output_identifiers: List[str] = []
        query_answers: List[QueryAnswerType] = []
        non_stringified_output_identifiers: List[Union[int, str, Dict[str, int]]] = []
        
        # Create query objects
        query_objs = []
        for query in queries:
            # Handle both dict and QueryType
            if isinstance(query, dict):
                query_dict = query
            else:
                query_dict = dict(query)
            
            # Convert string enums in query if needed
            if "designator" in query_dict and isinstance(query_dict["designator"], str):
                query_dict["designator"] = DesignatorType(query_dict["designator"])
            if "category" in query_dict and isinstance(query_dict["category"], str):
                query_dict["category"] = CategoryType(query_dict["category"])
            
            query_options = {
                "segmentation": segmentation,
                "designator": query_dict.get("designator"),
                "category": query_dict.get("category"),
                "pitch": query_dict.get("pitch"),
                "sequence_length": sequence_length,
                "trajectory_id": query_dict.get("trajectory_id"),
                "vowel": query_dict.get("vowel"),
                "consonant": query_dict.get("consonant"),
                "pitch_sequence": query_dict.get("pitch_sequence"),
                "traj_id_sequence": query_dict.get("traj_id_sequence"),
                "section_top_level": query_dict.get("section_top_level"),
                "alap_section": query_dict.get("alap_section"),
                "comp_type": query_dict.get("comp_type"),
                "comp_sec_tempo": query_dict.get("comp_sec_tempo"),
                "tala": query_dict.get("tala"),
                "phrase_type": query_dict.get("phrase_type"),
                "elaboration_type": query_dict.get("elaboration_type"),
                "vocal_art_type": query_dict.get("vocal_art_type"),
                "inst_art_type": query_dict.get("inst_art_type"),
                "incidental": query_dict.get("incidental"),
                "min_dur": min_dur,
                "max_dur": max_dur,
                "instrument_idx": instrument_idx,
            }
            query_objs.append(Query(piece, query_options))
        
        if every:
            # Only select trajectories that are in all answers
            if query_objs:
                output_identifiers = query_objs[0].stringified_identifier[:]
                for answer in query_objs[1:]:
                    output_identifiers = [
                        id_str for id_str in output_identifiers 
                        if id_str in answer.stringified_identifier
                    ]
                
                # Get corresponding trajectories and answers
                idxs = [
                    query_objs[0].stringified_identifier.index(id_str) 
                    for id_str in output_identifiers
                ]
                output_trajectories = [query_objs[0].trajectories[idx] for idx in idxs]
                non_stringified_output_identifiers = [query_objs[0].identifier[idx] for idx in idxs]
                query_answers = [query_objs[0].query_answers[idx] for idx in idxs]
        else:
            # Select trajectories that are in any answer
            start_times = []
            seen_ids = set()
            
            for answer in query_objs:
                for s_idx, s_id in enumerate(answer.stringified_identifier):
                    if s_id not in seen_ids:
                        seen_ids.add(s_id)
                        output_identifiers.append(s_id)
                        output_trajectories.append(answer.trajectories[s_idx])
                        non_stringified_output_identifiers.append(answer.identifier[s_idx])
                        query_answers.append(answer.query_answers[s_idx])
                        start_times.append(answer.start_times[s_idx])
            
            # Sort by start times
            sort_idxs = sorted(range(len(start_times)), key=lambda i: start_times[i])
            output_trajectories = [output_trajectories[idx] for idx in sort_idxs]
            non_stringified_output_identifiers = [non_stringified_output_identifiers[idx] for idx in sort_idxs]
            query_answers = [query_answers[idx] for idx in sort_idxs]
        
        return output_trajectories, non_stringified_output_identifiers, query_answers



"""Sequence matching utilities for query analysis.

This module provides exact Python ports of the TypeScript sequence matching
functions used in the query system.
"""

from typing import List, Dict, Optional, Tuple


def find_sequence_indexes(sequence: List[int], longer_sequence: List[int]) -> List[int]:
    """Find all starting indexes where sequence appears in longer_sequence.
    
    This is an exact port of the TypeScript findSequenceIndexes function.
    
    Args:
        sequence: The sequence to search for
        longer_sequence: The sequence to search in
        
    Returns:
        List of starting indexes where the sequence is found
    """
    indexes: List[int] = []
    
    if not sequence or not longer_sequence:
        return indexes
        
    start_index = -1
    try:
        start_index = longer_sequence.index(sequence[0])
    except ValueError:
        return indexes
    
    if start_index != -1:
        # Check first potential match
        match = True
        for i in range(1, len(sequence)):
            if start_index + i >= len(longer_sequence) or longer_sequence[start_index + i] != sequence[i]:
                match = False
                break
        
        if match:
            indexes.append(start_index)
        
        # Check remaining potential matches
        for i in range(start_index + 1, len(longer_sequence)):
            if longer_sequence[i] == sequence[0]:
                match = True
                for j in range(1, len(sequence)):
                    if i + j >= len(longer_sequence) or longer_sequence[i + j] != sequence[j]:
                        match = False
                        break
                
                if match:
                    indexes.append(i)
    
    return indexes


def loose_sequence_indexes(
    sequence: List[int], 
    longer_sequence: List[int]
) -> Dict[str, Optional[int] | bool]:
    """Test if sequence appears loosely (in order but not necessarily consecutively) in longer_sequence.
    
    This is an exact port of the TypeScript testLooseSequenceIndexes function.
    
    Args:
        sequence: The sequence to search for
        longer_sequence: The sequence to search in
        
    Returns:
        Dictionary with keys:
        - 'truth': Whether the sequence was found
        - 'first_idx': Index of first element match (None if not found)
        - 'last_idx': Index of last element match (None if not found)
    """
    ct = 0
    out = False
    first_idx: Optional[int] = None
    last_idx: Optional[int] = None
    
    if not sequence or not longer_sequence:
        return {"truth": out, "first_idx": first_idx, "last_idx": last_idx}
    
    for i in range(len(longer_sequence)):
        el = longer_sequence[i]
        
        if ct < len(sequence) and el == sequence[ct]:
            if ct == 0:
                first_idx = i
            ct += 1
        
        if ct > len(sequence) - 1:
            out = True
            last_idx = i
            break
    
    return {"truth": out, "first_idx": first_idx, "last_idx": last_idx}


def split_trajs_by_silences(trajectories: List) -> List[List]:
    """Split trajectory list by silences (trajectory ID 12).
    
    This is a port of the TypeScript splitTrajsBySilences method.
    
    Args:
        trajectories: List of trajectory objects
        
    Returns:
        List of trajectory sublists split by silence markers
    """
    result = [[]]
    
    for traj in trajectories:
        # Check if trajectory has ID 12 (silence marker)
        traj_id = getattr(traj, 'id', None)
        if traj_id == 12:
            result.append([])
        else:
            if len(result) == 0:
                result.append([])
            result[-1].append(traj)
    
    # Filter out empty sublists
    return [traj_seq for traj_seq in result if len(traj_seq) > 0]